"""
Dataproc cluster utils
"""
import math
from typing import Optional, List

from analysis_runner import dataproc

from hailtop.batch.job import Job
from hailtop.batch import Batch

from larcoh import utils


DataprocCluster = dataproc.DataprocCluster


# GCP quota on number of cores
MAX_PRIMARY_WORKERS = 50


def add_job(
    b: Batch,
    script: str,
    job_name: str,
    num_workers: int,
    long: bool = False,
    highmem: bool = False,
    is_test: bool = False,
    phantomjs: bool = False,
    preemptible: bool = True,
    depends_on: Optional[List[Optional[Job]]] = None,
    autoscaling_policy: Optional[str] = None,
    max_age: Optional[str] = None,
    pyfiles: Optional[List[str]] = None,
    worker_boot_disk_size: Optional[int] = None,
    secondary_worker_boot_disk_size: Optional[int] = None,
) -> Job:
    """
    Wrapper around submitting a dataproc cluster job
    """
    if not max_age:
        max_age = '1h' if is_test else '24h'
        if long:
            max_age = '3h' if is_test else '48h'

    depends_on = depends_on or []
    depends_on = [j for j in depends_on if j is not None]

    if preemptible:
        num_secondary_workers = num_workers
        # number of primary workers has to be 5-10% of the number of secondary workers:
        # see Tim's comment in https://discuss.hail.is/t/all-nodes-are-unhealthy/1764
        # using 8% here:
        num_primary_workers = int(math.ceil(num_secondary_workers * 0.08))
    else:
        num_secondary_workers = 0
        num_primary_workers = num_workers

    # 2 is the minimal number of primary workers for dataproc cluster:
    num_primary_workers = max(num_primary_workers, 2)

    # limiting the number of primary workers to avoid hitting GCP quota:
    num_primary_workers = min(num_primary_workers, MAX_PRIMARY_WORKERS)

    return dataproc.hail_dataproc_job(
        b,
        script=script,
        job_name=job_name,
        max_age=max_age,
        packages=utils.DATAPROC_PACKAGES,
        num_secondary_workers=num_secondary_workers,
        num_workers=num_primary_workers,
        autoscaling_policy=autoscaling_policy,
        depends_on=depends_on,
        init=['gs://cpg-reference/hail_dataproc/install_phantomjs.sh']
        if phantomjs
        else [],
        worker_machine_type='n1-highmem-8' if highmem else 'n1-standard-8',
        pyfiles=pyfiles,
        worker_boot_disk_size=worker_boot_disk_size,
        secondary_worker_boot_disk_size=secondary_worker_boot_disk_size,
    )


def get_cluster(
    b: Batch,
    name,
    num_workers: int,
    long: bool = False,
    highmem: bool = False,
    is_test: bool = False,
    phantomjs: bool = False,
    preemptible: bool = True,
    depends_on: Optional[List[Optional[Job]]] = None,
    autoscaling_policy: Optional[str] = None,
    max_age: Optional[str] = None,
) -> dataproc.DataprocCluster:
    """
    Get or create a Dataproc cluster by name
    """
    if not max_age:
        max_age = '1h' if is_test else '24h'
        if long:
            max_age = '3h' if is_test else '48h'

    depends_on = depends_on or []
    depends_on = [j for j in depends_on if j is not None]

    return dataproc.setup_dataproc(
        b,
        max_age=max_age,
        packages=utils.DATAPROC_PACKAGES,
        num_secondary_workers=num_workers if preemptible else 0,
        num_workers=2 if preemptible else num_workers,
        cluster_name=name,
        depends_on=depends_on,
        init=['gs://cpg-reference/hail_dataproc/install_phantomjs.sh']
        if phantomjs
        else [],
        autoscaling_policy=autoscaling_policy,
        worker_machine_type='n1-highmem-8' if highmem else 'n1-standard-8',
    )



#!/usr/bin/env python

"""
Plot ancestry PCA analysis results
"""

import logging
from collections import Counter
from typing import List, Iterable, Optional

import click
import pandas as pd
import numpy as np
import hail as hl
from bokeh.resources import CDN
from bokeh.embed import file_html
from bokeh.transform import factor_cmap, factor_mark
from bokeh.plotting import ColumnDataSource, figure
from bokeh.palettes import turbo  # flake: disable=F401
from bokeh.models import CategoricalColorMapper, HoverTool

from joint_calling import utils, resources
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--meta-tsv',
    'meta_tsv_path',
    callback=utils.get_validation_callback(ext='tsv', must_exist=True),
    required=True,
)
@click.option(
    '--eigenvalues-ht',
    'eigenvalues_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--scores-ht',
    'scores_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--loadings-ht',
    'loadings_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--inferred-pop-ht',
    'inferred_pop_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option('--out-path-pattern', 'out_path_pattern')
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    meta_tsv_path: str,
    eigenvalues_ht_path: str,
    scores_ht_path: str,
    loadings_ht_path: str,
    inferred_pop_ht_path: str,
    out_path_pattern: Optional[str],
    hail_billing: str,  # pylint: disable=unused-argument
):
    local_tmp_dir = utils.init_hail(__file__)

    meta_ht = utils.parse_input_metadata(meta_tsv_path, local_tmp_dir)

    produce_plots(
        eigenvalues_ht_path=eigenvalues_ht_path,
        scores_ht_path=scores_ht_path,
        loadings_ht_path=loadings_ht_path,
        inferred_pop_ht_path=inferred_pop_ht_path,
        meta_ht=meta_ht,
        out_path_pattern=out_path_pattern,
    )


def summarize_projects_by_pop(ht):
    """
    Make labels that split samples for a population by project
    """
    cnts_by_pop = dict()
    data = ht.select('pop', 'project').collect()
    for d in data:
        if d.pop not in cnts_by_pop:
            cnts_by_pop[d.pop] = dict()
        if d.project not in cnts_by_pop[d.pop]:
            cnts_by_pop[d.pop][d.project] = 0
        cnts_by_pop[d.pop][d.project] += 1
    
    summary_by_pop = {
        pop: ', '.join(f'{project}: {cnt}' for project, cnt in sorted(proj_d.items())) 
        for pop, proj_d in cnts_by_pop.items()
    }
    summary_by_pop = {k: v for k, v in summary_by_pop.items() if k}
    summary_by_pop = hl.literal(summary_by_pop)
    return summary_by_pop


BG_LABEL = 'Provided ancestry (1KG+HGDP)'
FG_LABEL = 'Inferred ancestry'


def produce_plots(
    eigenvalues_ht_path: str,
    scores_ht_path: str,
    loadings_ht_path: str,
    inferred_pop_ht_path: str,
    meta_ht: hl.Table,
    out_path_pattern: Optional[str] = None,
    number_of_pcs: Optional[int] = None,
):
    """
    Generate plots in HTML format, write for each PC (of n_pcs) and
    scope ("project", "population") plus for loadings) into
    file paths defined by `out_path_pattern`.
    """
    scores_ht = hl.read_table(scores_ht_path)
    inferred_pop_ht = hl.read_table(inferred_pop_ht_path)

    scores_ht = scores_ht.annotate(
        pop=inferred_pop_ht[scores_ht.s].pop,
        is_training=inferred_pop_ht[scores_ht.s].is_training,
        project=meta_ht[scores_ht.s].project
    ).cache()

    def key_by_external_id(ht, meta_ht=None):
        """
        Assuming ht.s is a CPG id, replaces it with external ID, 
        assuming it's defined in meta_ht.external_id
        """
        if meta_ht is None:
            meta_ht = ht
        ht = ht.annotate(old_s=ht.s).key_by('old_s')
        ht = (
            ht.annotate(
                s=hl.if_else(
                    hl.is_defined(meta_ht[ht.old_s]),
                    meta_ht[ht.old_s].external_id,
                    ht.old_s,
                )
            )
            .key_by('s')
            .drop('old_s')
        )
        return ht
    
    ht = key_by_external_id(scores_ht, meta_ht)

    pop_full_names = hl.literal({
        'nfe': 'Non-Finnish European', 
        'fin': 'Finnish',
        'mid': 'Middle Eastern',
        'oth': 'Other',
        'afr': 'African/African American',
        'eas': 'East Asian',
        'sas': 'South Asian',
        'amr': 'Latino/Admixed American',
    })
    ht = ht.annotate(pop=pop_full_names[ht.pop])
    ht = ht.cache()

    eigenvalues = hl.read_table(eigenvalues_ht_path).f0.collect()
    eigenvalues_df = pd.to_numeric(eigenvalues)
    variance = np.divide(eigenvalues_df[1:], float(eigenvalues_df.sum())) * 100
    variance = variance.round(2)
    number_of_pcs = number_of_pcs or len(eigenvalues_df) - 1

    plots = []

    sample_names = ht.s.collect()
    projects = ht.project.collect()
    is_training = ht.is_training.collect()
    
    for scope, title, labels in [
        ('project', 'Project', projects),
        ('population', 'Population', ht.pop.collect()),
    ]:
        plots.extend(
            _plot_pca(
                scope=scope,
                title=title,
                labels=labels,
                number_of_pcs=number_of_pcs,
                variance=variance,
                ht=ht,
                projects=projects,
                is_training=is_training,
                sample_names=sample_names,
                out_path_pattern=out_path_pattern,
            )
        )

    plots.extend(
        _plot_loadings(
            number_of_pcs, loadings_ht_path, out_path_pattern=out_path_pattern
        )
    )

    return plots


def _plot_pca(
    scope,
    title,
    labels,
    number_of_pcs,
    variance,
    ht,
    projects,
    sample_names,
    is_training,
    out_path_pattern=None,
):
    
    cntr: Counter = Counter(labels)
    labels = [f'{x} ({cntr[x]})' for x in labels]

    unique_labels = list(Counter(labels).keys())
    tooltips = [('labels', '@label'), ('samples', '@samples')]
    plots = []
    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title=title,
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
            width=1000,
        )
        source = ColumnDataSource(
            dict(
                x=ht.scores[pc1].collect(),
                y=ht.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
                project=projects,
                is_training=[
                    {True: BG_LABEL, False: FG_LABEL}.get(v) 
                    for v in is_training
                ],
            )
        )
        plot.scatter(
            'x',
            'y',
            alpha=0.5,
            marker=factor_mark('is_training', ['cross', 'circle'], [BG_LABEL, FG_LABEL]),
            source=source,
            size=4,
            color=factor_cmap('label', ['#1b9e77', '#d95f02'], unique_labels),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plots.append(plot)
        if out_path_pattern:
            html = file_html(plot, CDN, title)
            plot_filename_html = out_path_pattern.format(
                scope=scope, pci=pc2, ext='html'
            )
            with hl.hadoop_open(plot_filename_html, 'w') as f:
                f.write(html)
    return plots


def _plot_loadings(number_of_pcs, loadings_ht_path, out_path_pattern=None):
    loadings_ht = hl.read_table(loadings_ht_path)
    plots = []
    gtf_ht = hl.experimental.import_gtf(
        resources.GENCODE_GTF,
        reference_genome='GRCh38',
        skip_invalid_contigs=True,
        min_partitions=12,
        force_bgz=True,
    )
    for i in range(number_of_pcs - 1):
        pc = i + 1
        plot = manhattan_loadings(
            iteration=i,
            gtf=gtf_ht,
            loadings=loadings_ht,
            title='Loadings of PC ' + str(pc),
            collect_all=True,
        )
        plots.append(plot)
        if out_path_pattern:
            html = file_html(plot, CDN, 'my plot')
            plot_filename_html = out_path_pattern.format(
                scope='loadings', pci=pc, ext='html'
            )
            with hl.hadoop_open(plot_filename_html, 'w') as f:
                f.write(html)
    return plots


def manhattan_loadings(
    iteration,
    gtf,
    loadings,
    title=None,
    size=4,
    hover_fields=None,
    collect_all=False,
    n_divisions=500,
):
    """modify hail manhattan plot"""
    palette = [
        '#1f77b4',
        '#ff7f0e',
        '#2ca02c',
        '#d62728',
        '#9467bd',
        '#8c564b',
        '#e377c2',
        '#7f7f7f',
        '#bcbd22',
        '#17becf',
    ]

    # add gene names, p-values, and locus info
    loadings = loadings.annotate(gene_names=gtf[loadings.locus].gene_name)
    pvals = hl.abs(loadings.loadings[iteration])
    locus = loadings.locus

    if hover_fields is None:
        hover_fields = {}

    hover_fields['locus'] = hl.str(locus)
    hover_fields['gene'] = hl.str(loadings.gene_names)

    source_pd = (
        hl.plot.plots._collect_scatter_plot_data(  # pylint: disable=protected-access
            ('_global_locus', locus.global_position()),
            ('_pval', pvals),
            fields=hover_fields,
            n_divisions=None if collect_all else n_divisions,
        )
    )
    source_pd['p_value'] = source_pd['_pval']
    source_pd['_contig'] = [locus.split(':')[0] for locus in source_pd['locus']]

    observed_contigs = set(source_pd['_contig'])
    ref = locus.dtype.reference_genome
    observed_contigs = [
        contig for contig in ref.contigs.copy() if contig in observed_contigs
    ]

    contig_ticks = [
        ref._contig_global_position(contig)  # pylint: disable=protected-access
        + ref.contig_length(contig) // 2
        for contig in observed_contigs
    ]
    color_mapper = CategoricalColorMapper(
        factors=ref.contigs, palette=palette[:2] * int((len(ref.contigs) + 1) / 2)
    )

    p = figure(
        title=title, x_axis_label='Chromosome', y_axis_label='Loadings', width=1000
    )
    (
        p,
        _,
        legend,
        _,
        _,
        _,
    ) = hl.plot.plots._get_scatter_plot_elements(  # pylint: disable=protected-access
        p,
        source_pd,
        x_col='_global_locus',
        y_col='_pval',
        label_cols=['_contig'],
        colors={'_contig': color_mapper},
        size=size,
    )
    legend.visible = False
    p.xaxis.ticker = contig_ticks
    p.xaxis.major_label_overrides = dict(zip(contig_ticks, observed_contigs))
    p.select_one(HoverTool).tooltips = [
        t for t in p.select_one(HoverTool).tooltips if not t[0].startswith('_')
    ]

    return p


def remove_duplicates(x: Iterable) -> List:
    """
    Removes duplicates from a list, keeps order
    """
    return list(dict.fromkeys(x))


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Pipeline stage target and subclasses: Cohort, Dataset, Sample.
Plus Sample properties: PedigreeInfo, Sex.
"""
import hashlib
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional
import pandas as pd

from cpg_utils.hail_batch import dataset_path, web_url
from cpg_utils.config import get_config
from cpg_utils import Path, to_path

from larcoh.filetypes import (
    AlignmentInput,
    CramPath,
    GvcfPath,
    SequencingType,
    FastqPairs,
)

logger = logging.getLogger(__file__)


class Target:
    """
    Defines a target that stage can act upon. Classes like Sample, Dataset, Pipeline
    extend this class.
    """

    def __init__(self):
        # Whether to process even if outputs exist:
        self.forced: bool = False
        # If not set, exclude from the pipeline:
        self.active: bool = True

    def get_samples(self, only_active: bool = True) -> list['Sample']:
        """
        Get flat list of all samples corresponding to this target.
        """
        raise NotImplementedError

    def get_sample_ids(self, only_active: bool = True) -> list[str]:
        """
        Get flat list of all sample IDs corresponding to this target.
        """
        return [s.id for s in self.get_samples(only_active=only_active)]

    def alignment_inputs_hash(self) -> str:
        """
        Unique hash string of sample alignment inputs. Useful to decide
        whether the analysis on the target needs to be rerun.
        """
        s = ' '.join(
            sorted(
                [
                    ' '.join(
                        sorted(
                            str(alignment_input)
                            for alignment_input in s.alignment_input_by_seq_type.values()
                        )
                    )
                    for s in self.get_samples()
                    if s.alignment_input_by_seq_type
                ]
            )
        )
        h = hashlib.sha256(s.encode()).hexdigest()[:38]
        return f'{h}_{len(self.get_sample_ids())}'

    @property
    def target_id(self) -> str:
        """
        ID should be unique across target of all levels.

        We are raising NotImplementedError instead of making it an abstract class,
        because mypy is not happy about binding TypeVar to abstract classes, see:
        https://stackoverflow.com/questions/48349054/how-do-you-annotate-the-type-of
        -an-abstract-class-with-mypy

        Specifically,
        ```
        TypeVar('TargetT', bound=Target)
        ```
        Will raise:
        ```
        Only concrete class can be given where "Type[Target]" is expected
        ```
        """
        raise NotImplementedError

    def get_job_attrs(self) -> dict:
        """
        Attributes for Hail Batch job.
        """
        raise NotImplementedError

    def get_job_prefix(self) -> str:
        """
        Prefix job names.
        """
        raise NotImplementedError

    def rich_id_map(self) -> dict[str, str]:
        """
        Map if internal IDs to participant or external IDs, if the latter is provided.
        """
        return {s.id: s.rich_id for s in self.get_samples() if s.participant_id != s.id}


class Cohort(Target):
    """
    Represents a "cohort" target - all samples from all datasets in the pipeline.
    Analysis dataset name is required and will be used as the default name for the
    cohort.
    """

    def __init__(
        self,
        analysis_dataset_name: str,
        sequencing_type: SequencingType,
        name: str | None = None,
    ):
        super().__init__()
        self.name = name or analysis_dataset_name
        self.analysis_dataset = Dataset(name=analysis_dataset_name, cohort=self)
        self.sequencing_type = sequencing_type
        self._datasets_by_name: dict[str, Dataset] = {}

    def __repr__(self):
        return f'Cohort("{self.name}", {len(self.get_datasets())} datasets)'

    @property
    def target_id(self) -> str:
        """Unique target ID"""
        return self.name

    def get_datasets(self, only_active: bool = True) -> list['Dataset']:
        """
        Gets list of all datasets.
        Include only "active" datasets (unless only_active is False)
        """
        datasets = [ds for k, ds in self._datasets_by_name.items()]
        if only_active:
            datasets = [ds for ds in datasets if ds.active and ds.get_samples()]
        return datasets

    def get_dataset_by_name(
        self, name: str, only_active: bool = True
    ) -> Optional['Dataset']:
        """
        Get dataset by name.
        Include only "active" datasets (unless only_active is False)
        """
        ds_by_name = {d.name: d for d in self.get_datasets(only_active)}
        return ds_by_name.get(name)

    def get_samples(self, only_active: bool = True) -> list['Sample']:
        """
        Gets a flat list of all samples from all datasets.
        Include only "active" samples (unless only_active is False)
        """
        all_samples = []
        for ds in self.get_datasets(only_active=False):
            all_samples.extend(ds.get_samples(only_active=only_active))
        return all_samples

    def add_dataset(self, dataset: 'Dataset') -> 'Dataset':
        """
        Add existing dataset into the cohort.
        """
        dataset.cohort = self
        if dataset.name in self._datasets_by_name:
            logger.debug(f'Dataset {dataset.name} already exists in the cohort')
            return dataset
        self._datasets_by_name[dataset.name] = dataset
        return dataset

    def create_dataset(
        self,
        name: str,
    ) -> 'Dataset':
        """
        Create a dataset and add it to the cohort.
        """
        if name in self._datasets_by_name:
            logger.debug(f'Dataset {name} already exists in the cohort')
            return self._datasets_by_name[name]

        if name == self.analysis_dataset.name:
            ds = self.analysis_dataset
        else:
            ds = Dataset(name=name, cohort=self)

        self._datasets_by_name[ds.name] = ds
        return ds

    def get_job_attrs(self) -> dict:
        """
        Attributes for Hail Batch job.
        """
        return {
            'samples': self.get_sample_ids(),
            'datasets': [d.name for d in self.get_datasets()],
        }

    def get_job_prefix(self) -> str:
        """
        Prefix job names.
        """
        return ''

    def to_tsv(self) -> str:
        """
        Export to a parsable TSV file
        """
        tsv_path = self.analysis_dataset.tmp_prefix() / 'samples.tsv'
        df = pd.DataFrame(
            {
                's': s.id,
                'gvcf': s.gvcf or '-',
                'sex': s.meta.get('sex') or '-',
                'continental_pop': s.meta.get('continental_pop') or '-',
                'subcontinental_pop': s.meta.get('subcontinental_pop') or '-',
            }
            for s in self.get_samples()
        ).set_index('s', drop=False)
        with to_path(tsv_path).open('w') as f:
            df.to_csv(f, index=False, sep='\t', na_rep='NA')
        return tsv_path


class Dataset(Target):
    """
    Represents a CPG dataset.

    Each `dataset` at the CPG corresponds to
    * a GCP project: https://github.com/populationgenomics/team-docs/tree/main/storage_policies
    * a Pulumi stack: https://github.com/populationgenomics/analysis-runner/tree/main/stack
    * a metamist project
    """

    def __init__(
        self,
        name: str,
        cohort: Cohort | None = None,
    ):
        super().__init__()
        self._sample_by_id: dict[str, Sample] = {}
        self.name = name
        self.cohort = cohort
        self.active = True

    @staticmethod
    def create(name: str) -> 'Dataset':
        """
        Create a dataset.
        """
        return Dataset(name=name)

    @property
    def target_id(self) -> str:
        """Unique target ID"""
        return self.name

    def __repr__(self):
        return f'Dataset("{self.name}", {len(self.get_samples())} samples)'

    def __str__(self):
        return f'{self.name} ({len(self.get_samples())} samples)'

    def _seq_type_subdir(self) -> str:
        """
        Subdirectory parametrised by sequencing type. For genomes, we don't
        prefix at all.
        """
        return (
            ''
            if not self.cohort or self.cohort.sequencing_type == SequencingType.GENOME
            else self.cohort.sequencing_type.value
        )

    def prefix(self, **kwargs) -> Path:
        """
        The primary storage path.
        """
        return to_path(
            dataset_path(
                self._seq_type_subdir(),
                dataset=self.name,
                **kwargs,
            )
        )

    def tmp_prefix(self, **kwargs) -> Path:
        """
        Storage path for temporary files.
        """
        return to_path(
            dataset_path(
                self._seq_type_subdir(),
                dataset=self.name,
                category='tmp',
                **kwargs,
            )
        )

    def web_prefix(self, **kwargs) -> Path:
        """
        Path for files served by an HTTP server Matches corresponding URLs returns by
        self.web_url() URLs.
        """
        return to_path(
            dataset_path(
                self._seq_type_subdir(),
                dataset=self.name,
                category='web',
                **kwargs,
            )
        )

    def web_url(self, **kwargs) -> str | None:
        """
        URLs matching self.storage_web_path() files serverd by an HTTP server.
        """
        return web_url(
            self._seq_type_subdir(),
            dataset=self.name,
            **kwargs,
        )

    def add_sample(
        self,
        id: str,  # pylint: disable=redefined-builtin
        external_id: str | None = None,
        participant_id: str | None = None,
        meta: dict | None = None,
        sex: Optional['Sex'] = None,
        pedigree: Optional['PedigreeInfo'] = None,
        alignment_input_by_seq_type: dict[SequencingType, AlignmentInput] | None = None,
    ) -> 'Sample':
        """
        Create a new sample and add it to the dataset.
        """
        if id in self._sample_by_id:
            logger.debug(f'Sample {id} already exists in the dataset {self.name}')
            return self._sample_by_id[id]

        force_samples = get_config()['workflow'].get('force_samples', set())
        forced = (
            id in force_samples
            or external_id in force_samples
            or participant_id in force_samples
        )

        s = Sample(
            id=id,
            dataset=self,
            external_id=external_id,
            participant_id=participant_id,
            meta=meta,
            sex=sex,
            pedigree=pedigree,
            alignment_input_by_seq_type=alignment_input_by_seq_type,
            forced=forced,
        )
        self._sample_by_id[id] = s
        return s

    def get_samples(self, only_active: bool = True) -> list['Sample']:
        """
        Get dataset's samples. Include only "active" samples, unless only_active=False
        """
        return [
            s for sid, s in self._sample_by_id.items() if (s.active or not only_active)
        ]

    def get_job_attrs(self) -> dict:
        """
        Attributes for Hail Batch job.
        """
        return {
            'dataset': self.name,
            'samples': self.get_sample_ids(),
        }

    def get_job_prefix(self) -> str:
        """
        Prefix job names.
        """
        return f'{self.name}: '

    def make_ped_file(self, tmp_bucket: Path | None = None) -> Path:
        """
        Create a PED file for all samples
        """
        datas = []
        for sample in self.get_samples():
            if sample.pedigree:
                datas.append(sample.pedigree.get_ped_dict())
        df = pd.DataFrame(datas)

        ped_path = (tmp_bucket or self.tmp_prefix()) / f'{self.name}.ped'
        with ped_path.open('w') as fp:
            df.to_csv(fp, sep='\t', index=False)

        return ped_path


class Sex(Enum):
    """
    Sex as in PED format
    """

    UNKNOWN = 0
    MALE = 1
    FEMALE = 2

    @staticmethod
    def parse(sex: str | None) -> 'Sex':
        """
        Parse a string into a Sex object.
        """
        if sex:
            if sex.lower() in ('m', 'male', '1'):
                return Sex.MALE
            if sex.lower() in ('f', 'female', '2'):
                return Sex.FEMALE
            if sex.lower() in ('u', 'unknown', '0'):
                return Sex.UNKNOWN
            raise ValueError(f'Unrecognised sex value {sex}')
        return Sex.UNKNOWN

    def __str__(self):
        return self.name


class Sample(Target):
    """
    Represents a Sample.
    """

    def __init__(
        self,
        id: str,  # pylint: disable=redefined-builtin
        dataset: 'Dataset',  # type: ignore  # noqa: F821
        external_id: str | None = None,
        participant_id: str | None = None,
        meta: dict | None = None,
        sex: Sex | None = None,
        pedigree: Optional['PedigreeInfo'] = None,
        alignment_input_by_seq_type: dict[SequencingType, AlignmentInput] | None = None,
        forced: bool = False,
    ):
        super().__init__()
        self.id = id
        self._external_id = external_id
        self.dataset = dataset
        self._participant_id = participant_id
        self.meta: dict = meta or dict()
        self.pedigree: PedigreeInfo | None = pedigree
        if sex:
            self.pedigree = PedigreeInfo(
                sample=self,
                fam_id=self.participant_id,
                sex=sex,
            )
        self.alignment_input_by_seq_type: dict[SequencingType, AlignmentInput] = (
            alignment_input_by_seq_type or dict()
        )
        self.sequencing_meta_by_type: dict[SequencingType, dict] = {}
        self.forced = forced
        self.active = True
        # Only set if the file exists / found in Metamist:
        self.gvcf: GvcfPath | None = None
        self.cram: CramPath | None = None

    def __repr__(self):
        values = {
            'participant': self._participant_id if self._participant_id else '',
            'forced': str(self.forced),
            'active': str(self.active),
            'meta': str(self.meta),
            'alignment_inputs': ','.join(
                [
                    f'{seq_t.value}: {al_inp}'
                    for seq_t, al_inp in self.alignment_input_by_seq_type.items()
                ]
            ),
            'pedigree': self.pedigree if self.pedigree else '',
        }
        retval = f'Sample({self.dataset.name}/{self.id}'
        if self._external_id:
            retval += f'|{self._external_id}'
        return retval + ''.join(f', {k}={v}' for k, v in values.items())

    def __str__(self):
        ai_tag = ''
        for seq_type, alignment_input in self.alignment_input_by_seq_type.items():
            ai_tag += f'|SEQ={seq_type.value}:'
            if isinstance(alignment_input, CramPath):
                if alignment_input.is_bam:
                    ai_tag += 'CRAM'
                else:
                    ai_tag += 'BAM'
            else:
                assert isinstance(alignment_input, FastqPairs)
                ai_tag += f'{len(alignment_input)}FQS'

        ext_id = f'|{self._external_id}' if self._external_id else ''
        return f'Sample({self.dataset.name}/{self.id}{ext_id}{ai_tag})'

    @property
    def participant_id(self) -> str:
        """
        Get ID of participant corresponding to this sample,
        or substitute it with external ID.
        """
        return self._participant_id or self.external_id

    @participant_id.setter
    def participant_id(self, val: str):
        """
        Set participant ID.
        """
        self._participant_id = val

    @property
    def external_id(self) -> str:
        """
        Get external sample ID, or substitute it with the internal ID.
        """
        return self._external_id or self.id

    @property
    def rich_id(self) -> str:
        """
        ID for reporting purposes: composed of internal as well as external
        or participant IDs.
        """
        return self.id + '|' + self.participant_id

    def get_ped_dict(self, use_participant_id: bool = False) -> dict[str, str]:
        """
        Returns a dictionary of pedigree fields for this sample, corresponding
        a PED file entry.
        """
        if self.pedigree:
            return self.pedigree.get_ped_dict(use_participant_id)
        return {
            'Family.ID': self.participant_id if use_participant_id else self.id,
            'Individual.ID': self.participant_id if use_participant_id else self.id,
            'Father.ID': '0',
            'Mother.ID': '0',
            'Sex': '0',
            'Phenotype': '0',
        }

    def make_cram_path(self) -> CramPath:
        """
        Path to a CRAM file. Not checking its existence here.
        """
        return CramPath(self.dataset.prefix() / 'cram' / f'{self.id}.cram')

    def make_gvcf_path(self, access_level: str | None = None) -> GvcfPath:
        """
        Path to a GVCF file. Not checking its existence here.
        """
        return GvcfPath(
            self.dataset.prefix(access_level=access_level)
            / 'gvcf'
            / f'{self.id}.g.vcf.gz'
        )

    @property
    def target_id(self) -> str:
        """Unique target ID"""
        return self.id

    def get_samples(self, only_active: bool = True) -> list['Sample']:
        """
        Implementing the abstract method.
        """
        if only_active and not self.active:
            return []
        return [self]

    def get_job_attrs(self) -> dict:
        """
        Attributes for Hail Batch job.
        """
        return {
            'dataset': self.dataset.name,
            'sample': self.id,
        }

    def get_job_prefix(self) -> str:
        """
        Prefix job names.
        """
        return f'{self.dataset.name}/{self.id}: '


@dataclass
class PedigreeInfo:
    """
    Pedigree relationships with other samples in the cohort, and other PED data
    """

    sample: Sample
    sex: Sex
    fam_id: str | None = None
    phenotype: str | None = None
    dad: Sample | None = None
    mom: Sample | None = None

    def get_ped_dict(self, use_participant_id: bool = False) -> dict:
        """
        Returns a dictionary of pedigree fields for this sample, corresponding
        a PED file entry.
        """

        def _get_id(_s: Sample | None):
            if _s is None:
                return '0'
            if use_participant_id:
                return _s.participant_id
            return _s.id

        return {
            'Family.ID': self.fam_id or self.sample.participant_id,
            'Individual.ID': _get_id(self.sample),
            'Father.ID': _get_id(self.dad),
            'Mother.ID': _get_id(self.mom),
            'Sex': str(self.sex.value),
            'Phenotype': self.phenotype,
        }



"""
Utility functions and constants.
"""
import collections
import os
import re
import unicodedata
import logging
import string
import sys
import time
import traceback
from functools import lru_cache
from random import choices
from typing import cast, Callable

import hail as hl
import click

from cpg_utils import Path, to_path
from cpg_utils.config import get_config

logger = logging.getLogger(__file__)


# Packages to install on a dataproc cluster, to use with the dataproc wrapper.
DATAPROC_PACKAGES = [
    'cpg_utils',
    'larcoh',
    'cpg_gnomad',  # github.com/populationgenomics/gnomad_methods
    'seqr_loader==1.2.5',  # hail-elasticsearch-pipelines
    'elasticsearch==8.1.1',
    'cpg_utils',
    'click',
    'google',
    'fsspec',
    'sklearn',
    'gcloud',
    'selenium',
]


@lru_cache
def exists(
    path: Path | str, verbose: bool = True, description: str | None = None
) -> bool:
    """
    Caching version of the existence check.
    The python code runtime happens entirely during the pipeline submission,
    without waiting for it to finish, so there is no expectation that object
    existence status would change during the runtime. This, this function uses
    `@lru_cache` to make sure that object existence is checked only once.
    """
    return exists_not_cached(path, verbose)


def exists_not_cached(
    path: Path | str, verbose: bool = True, description: str | None = None
) -> bool:
    """
    Check if the object exists, where the object can be:
        * local file
        * local directory
        * cloud object
        * cloud URL representing a *.mt or *.ht Hail data,
          in which case it will check for the existence of a
          *.mt/_SUCCESS or *.ht/_SUCCESS file.
    @param path: path to the file/directory/object/mt/ht
    @param verbose: print on each check
    @param description: optional string to print in verbose mode
    @return: True if the object exists
    """
    extra_msg = '' if not description else f' ({description})'

    path = cast(Path, to_path(path))

    # rstrip to ".mt/" -> ".mt"
    if any(str(path).rstrip('/').endswith(f'.{suf}') for suf in ['mt', 'ht']):
        path = path / '_SUCCESS'

    if verbose:
        # noinspection PyBroadException
        try:
            res = path.exists()
        except BaseException:
            traceback.print_exc()
            logger.error(f'Failed checking {path}' + extra_msg)
            sys.exit(1)
        logger.debug(
            f'Checked {path} [' + ('exists' if res else 'missing') + ']' + extra_msg
        )
        return res
    return path.exists()


def can_reuse(path: list[Path] | Path | str | None) -> bool:
    """
    Checks if `fpath` is good to reuse in the analysis: it exists
    and `overwrite` is False.

    If `fpath` is a collection, it requires all files in it to exist.
    """
    if get_config()['workflow'].get('overwrite', False):
        return False

    if not path:
        return False

    if isinstance(path, list):
        return all(can_reuse(fp) for fp in path)

    if not exists(path):
        return False

    logger.debug(f'Reusing existing {path}. Use --overwrite to overwrite')
    return True


def timestamp(rand_suffix_len: int = 5) -> str:
    """
    Generate a timestamp string. If `rand_suffix_len` is set, adds a short random
    string of this length for uniqueness.
    """
    result = time.strftime('%Y_%m%d_%H%M')
    if rand_suffix_len:
        rand_bit = ''.join(
            choices(string.ascii_uppercase + string.digits, k=rand_suffix_len)
        )
        result += f'_{rand_bit}'
    return result


def slugify(line: str):
    """
    Slugify a string.

    Example:
    >>> slugify(u"Héllø Wörld")
    u"hello-world"
    """

    return re.sub(
        r'[-\s]+',
        '-',
        re.sub(
            r'[^\w\s-]',
            '',
            unicodedata.normalize('NFKD', line).encode('ascii', 'ignore').decode(),
        )
        .strip()
        .lower(),
    )


def check_duplicates(iterable):
    """
    Throws error if input list contains repeated items.
    """
    duplicates = [
        item for item, count in collections.Counter(iterable).items() if count > 1
    ]
    if duplicates:
        raise ValueError(f'Found {len(duplicates)} duplicates: {duplicates}')
    return duplicates


def get_validation_callback(
    ext: str = None,
    must_exist: bool = False,
    accompanying_metadata_suffix: str = None,
) -> Callable:
    """
    Get callback for Click parameters validation
    :param ext: check that the path has the expected extension
    :param must_exist: check that the input file/object/directory exists
    :param accompanying_metadata_suffix: checks that a file at the same location but
    with a different suffix also exists (e.g. genomes.mt and genomes.metadata.ht)
    :return: a callback suitable for Click parameter initialization
    """

    def callback(_, param, value):
        if value is None:
            return None
        if ext:
            assert isinstance(value, str), value
            value = value.rstrip('/')
            if not value.endswith(f'.{ext}'):
                raise click.BadParameter(
                    f'The argument {param.name} is expected to have '
                    f'an extension .{ext}, got: {value}'
                )
        if must_exist:
            if not exists(value):
                raise click.BadParameter(f"{value} doesn't exist or incomplete")
            if accompanying_metadata_suffix:
                accompanying_metadata_fpath = (
                    os.path.splitext(value)[0] + accompanying_metadata_suffix
                )
                if not exists(accompanying_metadata_fpath):
                    raise click.BadParameter(
                        f"An accompanying file {accompanying_metadata_fpath} doesn't "
                        f'exist'
                    )
        return value

    return callback


def get_mt(
    mt_path: str,
    split: bool = False,
    hard_filtered_samples_to_remove_ht: hl.Table = None,
    meta_ht: hl.Table = None,
    add_meta: bool = False,
    release_only: bool = False,
    passing_sites_only: bool = False,
    unrelated_only: bool = False,
) -> hl.MatrixTable:
    """
    Wrapper function to get data with desired filtering and metadata annotations
    :param mt_path: path to the MatrixTable
    :param split:
        Split multiallelics and convert local-allele LGT/LA fields to GT.
        Note: this will perform a split on the MT rather than grab an already split MT
    :param hard_filtered_samples_to_remove_ht:
        table with samples to remove
        (only relevant after sample QC that produces a table with samples failed
        filtering)
    :param meta_ht: table with meta-information generated by sample QC
    :param add_meta: whether to add metadata to MT in 'meta' column
    :param release_only: whether to filter the MT to only samples available for
        release (can only be used with)
    :param passing_sites_only: whether to filter the MT to only variants with
        nothing in the filter field (e.g. passing soft filters)
    :param unrelated_only: remove related samples (keep one sample from a family)
    :return: MatrixTable with chosen annotations and filters
    """
    mt = hl.read_matrix_table(mt_path)

    if passing_sites_only:
        try:
            mt = mt.filter_rows(mt.filters.length() == 0)
        except AttributeError:
            pass

    if hard_filtered_samples_to_remove_ht is not None:
        mt = mt.filter_cols(
            hl.is_missing(hard_filtered_samples_to_remove_ht[mt.col_key])
        )

    if add_meta:
        assert meta_ht is not None
        mt = mt.annotate_cols(meta=meta_ht[mt.col_key])

        if release_only:
            mt = mt.filter_cols(mt.meta.release)

        if unrelated_only:
            mt = mt.filter_cols(~mt.meta.related)

    else:
        if release_only:
            assert meta_ht is not None
            mt = mt.filter_cols(meta_ht[mt.col_key].release)

        if unrelated_only:
            assert meta_ht is not None
            mt = mt.filter_cols(~meta_ht[mt.col_key].related)

    if split:
        mt = mt.annotate_rows(
            n_unsplit_alleles=hl.len(mt.alleles),
            mixed_site=(hl.len(mt.alleles) > 2)
            & hl.any(lambda a: hl.is_indel(mt.alleles[0], a), mt.alleles[1:])
            & hl.any(lambda a: hl.is_snp(mt.alleles[0], a), mt.alleles[1:]),
        )
        # Will use GT instead of LGT
        mt = hl.experimental.sparse_split_multi(mt, filter_changed_loci=True)

    return mt



import logging

from cpg_utils import to_path
from cpg_utils.config import get_config
from cpg_utils.hail_batch import dataset_path

from larcoh import (
    analysis_prefix,
    output_version,
    batch,
)
from larcoh.tasks import combiner
from larcoh.tasks.sample_qc import sample_qc

logger = logging.getLogger(__file__)


def main():
    ####################
    # VCF COMBINER

    raw_combined_mt_path = analysis_prefix / 'combiner' / f'{output_version}-raw.mt'
    if prev_version := get_config()['workflow'].get('raw_combined_mt_version'):
        prev_analysis_base = to_path(
            dataset_path(f'joint-calling/{prev_version}', category='analysis')
        )
        raw_combined_mt_path = (
            prev_analysis_base / 'combiner' / f'{prev_version}-raw.mt'
        )
    combiner_jobs = combiner.queue_jobs(out_mt_path=raw_combined_mt_path)

    ####################
    # SAMPLE QC

    sample_qc_prefix = analysis_prefix / 'qc'
    sex_ht_path = sample_qc_prefix / 'sex.ht'
    hard_filtered_samples_ht_path = sample_qc_prefix / 'hard_filtered_samples.ht'
    hail_sample_qc_ht_path = sample_qc_prefix / 'hail_sample_qc.ht'

    sample_qc(
        raw_combined_mt_path,
        hard_filtered_samples_ht_path,
        sex_ht_path,
        hail_sample_qc_ht_path,
        depends_on=combiner_jobs,
    )

    batch().run()

    # sample_qc_job, hard_filter_ht_path, meta_ht_path = add_sample_qc_jobs(
    #     b=b,
    #     mt_path=raw_combined_mt_path,
    #     samples_tsv_path=samples_tsv_path,
    #     sample_qc_bucket=join(analysis_bucket, 'sample_qc'),
    #     ancestry_bucket=join(analysis_bucket, 'ancestry'),
    #     tmp_bucket=tmp_bucket,
    #     analysis_bucket=analysis_bucket,
    #     relatedness_bucket=relatedness_bucket,
    #     release_related=release_related,
    #     web_bucket=web_bucket,
    #     filter_cutoffs_path=filter_cutoffs_path,
    #     overwrite=overwrite,
    #     scatter_count=scatter_count,
    #     combiner_job=combiner_job,
    #     billing_project=billing_project,
    #     is_test=output_namespace in ['test', 'tmp'],
    #     highmem_workers=highmem_workers,
    # )
    #
    # var_qc_job = add_variant_qc_jobs(
    #     b=b,
    #     work_bucket=join(analysis_bucket, 'variant_qc'),
    #     web_bucket=join(web_bucket, 'variant_qc'),
    #     raw_combined_mt_path=raw_combined_mt_path,
    #     hard_filter_ht_path=hard_filter_ht_path,
    #     meta_ht_path=meta_ht_path,
    #     out_filtered_combined_mt_path=filtered_combined_mt_path,
    #     sample_count=len(samples_df),
    #     ped_file=ped_fpath,
    #     overwrite=overwrite,
    #     vqsr_params_d=utils.get_filter_cutoffs(filter_cutoffs_path)['vqsr'],
    #     scatter_count=scatter_count,
    #     is_test=output_namespace in ['test', 'tmp'],
    #     depends_on=[combiner_job, sample_qc_job],
    #     highmem_workers=highmem_workers,
    # )
    #
    # job_name = 'Remove ref blocks'
    # if not utils.can_reuse(filtered_combined_noref_mt_path, overwrite):
    #     noref_mt_j = add_job(
    #         b,
    #         f'{utils.SCRIPTS_DIR}/make_noref_mt.py '
    #         f'--overwrite '
    #         f'--mt {filtered_combined_mt_path} '
    #         f'--out-mt {filtered_combined_noref_mt_path}',
    #         job_name=job_name,
    #         num_workers=scatter_count,
    #         depends_on=[var_qc_job],
    #     )
    # else:
    #     noref_mt_j = b.new_job(f'{job_name} [reuse]')
    #
    # if subset_projects:
    #     diff_projects = set(subset_projects) - set(input_projects)
    #     if diff_projects:
    #         raise click.BadParameter(
    #             f'--subset-project values should be a subset of --input-project '
    #             f'values. The following projects are not in input projects: '
    #             f'{diff_projects} '
    #         )
    #     subset_projects = list(set(subset_projects))
    #     subset_mt_path = f'{release_bucket}/mt/{output_version}-{"-".join(sorted(subset_projects))}.mt'
    #     job_name = f'Making subset MT for {", ".join(subset_projects)}'
    #     if overwrite or not utils.file_exists(subset_mt_path):
    #         add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/make_subset_mt.py '
    #             f'--mt {filtered_combined_noref_mt_path} ' +
    #             (''.join(f'--subset-project {p} ' for p in subset_projects)) +
    #             f'--out-mt {subset_mt_path}',
    #             job_name=job_name,
    #             is_test=output_namespace in ['test', 'tmp'],
    #             num_workers=scatter_count,
    #             depends_on=[noref_mt_j],
    #         )
    #     else:
    #         b.new_job(f'{job_name} [reuse]')
    #
    #     subset_vcf_path = f'{release_bucket}/vcf/{output_version}-{"-".join(subset_projects)}.vcf.bgz'
    #     job_name = f'Convert subset MT to VCF for {", ".join(subset_projects)}'
    #     if overwrite or not utils.file_exists(subset_vcf_path):
    #         add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/final_mt_to_vcf.py '
    #             f'--mt {subset_mt_path} ' +
    #             f'--out-vcf {subset_vcf_path}',
    #             job_name=job_name,
    #             is_test=output_namespace in ['test', 'tmp'],
    #             num_workers=scatter_count,
    #             preemptible=False,
    #             depends_on=[noref_mt_j],
    #         )
    #     else:
    #         b.new_job(f'{job_name} [reuse]')

    # def add_sample_qc_jobs(
    #     # b: hb.Batch,
    #     # mt_path: str,
    #     # samples_tsv_path: str,
    #     # sample_qc_bucket: str,
    #     # ancestry_bucket: str,
    #     # tmp_bucket: str,
    #     # analysis_bucket: str,
    #     # relatedness_bucket: str,
    #     # release_related: bool,
    #     # web_bucket: str,
    #     # filter_cutoffs_path: Optional[str],
    #     # overwrite: bool,
    #     # scatter_count: int,
    #     # combiner_job: Job,
    #     # billing_project: Optional[str] = None,
    #     # is_test: bool = False,
    #     # somalier_pairs_path: Optional[str] = None,
    #     # somalier_job: Optional[Job] = None,
    #     # highmem_workers: bool = False,
    # ):
    #     """
    #     Sub-workflow that adds sample-level QC, relatedness and ancestry analysis jobs
    #     using Hail query on the combined matrix table
    #     """
    #     job_name = 'Sample QC hard filters'
    #     hard_filtered_samples_ht_path = join(sample_qc_bucket, 'hard_filtered_samples.ht')
    #     sex_ht_path = join(sample_qc_bucket, 'sex.ht')
    #     hail_sample_qc_ht_path = join(sample_qc_bucket, 'hail_sample_qc.ht')
    #     if not can_reuse(
    #         [
    #             hard_filtered_samples_ht_path,
    #             sex_ht_path,
    #             hail_sample_qc_ht_path,
    #         ],
    #         overwrite,
    #     ):
    #         sample_qc_hardfilter_job = add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/sample_qc_hard_filters.py '
    #             f'--mt {mt_path} '
    #             f'--meta-tsv {samples_tsv_path} '
    #             f'{_make_filter_cutoff_param(filter_cutoffs_path, sample_qc_bucket)} '
    #             f'--out-hard-filtered-samples-ht {hard_filtered_samples_ht_path} '
    #             f'--out-sex-ht {sex_ht_path} '
    #             f'--out-hail-sample-qc-ht {hail_sample_qc_ht_path} '
    #             f'--tmp-bucket {tmp_bucket} '
    #             + (f'--overwrite ' if overwrite else '')
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             preemptible=False,
    #             is_test=is_test,
    #             phantomjs=True,
    #             highmem=highmem_workers,
    #             depends_on=[combiner_job],
    #         )
    #     else:
    #         sample_qc_hardfilter_job = b.new_job(f'{job_name} [reuse]')
    #
    #     job_name = 'Subset MT for PCA'
    #     mt_for_pca_path = join(ancestry_bucket, 'mt_for_pca.mt')
    #     if not can_reuse(mt_for_pca_path, overwrite):
    #         subset_for_pca_job = add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/sample_qc_subset_mt_for_pca.py '
    #             + (f'--overwrite ' if overwrite else '')
    #             + f'--mt {mt_path} '
    #             f'--meta-tsv {samples_tsv_path} '
    #             f'--out-mt {mt_for_pca_path} '
    #             + ('--is-test ' if is_test else '')
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             is_test=is_test,
    #             phantomjs=True,
    #             highmem=highmem_workers,
    #             depends_on=[combiner_job],
    #         )
    #     else:
    #         subset_for_pca_job = b.new_job(f'{job_name} [reuse]')
    #
    #     relatedness_ht_path = join(relatedness_bucket, 'relatedness.ht')
    #     if somalier_pairs_path:
    #         job_name = 'Somalier pairs to Hail table'
    #         if not can_reuse(relatedness_ht_path, overwrite):
    #             relatedness_j = add_job(
    #                 b,
    #                 f'{utils.SCRIPTS_DIR}/sample_qc_somalier_to_ht.py '
    #                 + (f'--overwrite ' if overwrite else '')
    #                 + f'--somalier-pairs-tsv {somalier_pairs_path} '
    #                 f'--out-relatedness-ht {relatedness_ht_path} '
    #                 + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #                 job_name=job_name,
    #                 num_workers=scatter_count,
    #                 is_test=is_test,
    #                 phantomjs=True,
    #                 highmem=highmem_workers,
    #                 depends_on=[somalier_job],
    #             )
    #         else:
    #             relatedness_j = b.new_job(f'{job_name} [reuse]')
    #     else:
    #         job_name = 'Run pc_relate'
    #         if not can_reuse(relatedness_ht_path, overwrite):
    #             relatedness_ht_path = join(relatedness_bucket, 'relatedness.ht')
    #             # PC relate needs non-preemptible workes, so requesting a new cluster
    #             relatedness_j = add_job(
    #                 b,
    #                 f'{utils.SCRIPTS_DIR}/sample_qc_pcrelate.py '
    #                 + (f'--overwrite ' if overwrite else '')
    #                 + f'--pca-mt {mt_for_pca_path} '
    #                 f'--out-relatedness-ht {relatedness_ht_path} '
    #                 f'--tmp-bucket {tmp_bucket} '
    #                 + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #                 job_name=job_name,
    #                 num_workers=scatter_count,
    #                 is_test=is_test,
    #                 # Spark would have problems shuffling on preemptible workers
    #                 # (throw SparkException: Job aborted due to stage failure: ShuffleMapStage)
    #                 # so we use num_workers instead of num_secondary_workers here
    #                 preemptible=False,
    #                 highmem=highmem_workers,
    #                 depends_on=[subset_for_pca_job],
    #             )
    #         else:
    #             relatedness_j = b.new_job(f'{job_name} [reuse]')
    #
    #     cutoffs_d = utils.get_filter_cutoffs(filter_cutoffs_path)
    #
    #     job_name = 'Sample QC flag related'
    #     intermediate_related_samples_to_drop_ht_path = join(
    #         relatedness_bucket, 'intermediate_related_samples_to_drop.ht'
    #     )
    #     if not can_reuse(intermediate_related_samples_to_drop_ht_path, overwrite):
    #         flag_related_job = add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/sample_qc_flag_related.py '
    #             f'--hard-filtered-samples-ht {hard_filtered_samples_ht_path} '
    #             f'--sex-ht {sex_ht_path} '
    #             f'--relatedness-ht {relatedness_ht_path} '
    #             f'--out-ht {intermediate_related_samples_to_drop_ht_path} '
    #             f'--tmp-bucket {tmp_bucket} '
    #             f'--max-kin {cutoffs_d["pca"]["max_kin"]} '
    #             + (f'--overwrite ' if overwrite else '')
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             is_test=is_test,
    #             phantomjs=True,
    #             highmem=highmem_workers,
    #             depends_on=[sample_qc_hardfilter_job, relatedness_j],
    #         )
    #     else:
    #         flag_related_job = b.new_job(f'{job_name} [reuse]')
    #
    #     mt_for_pca_union_hgdp_path = 'gs://cpg-prophecy-main-analysis/joint-calling/v1-1/ancestry-with-hgdp/mt_for_pca_union_hgdp.mt'
    #     ancestry_bucket = join(analysis_bucket, 'ancestry_with_hgdp')
    #     num_ancestry_pcs = 16
    #     job_name = 'PCA'
    #     ancestry_analysis_bucket = ancestry_bucket
    #     ancestry_web_bucket = join(web_bucket, 'ancestry_with_hgdp')
    #     eigenvalues_ht_path = join(ancestry_analysis_bucket, f'eigenvalues.ht')
    #     scores_ht_path = join(ancestry_analysis_bucket, f'scores.ht')
    #     loadings_ht_path = join(ancestry_analysis_bucket, f'loadings.ht')
    #     inferred_pop_ht_path = join(ancestry_bucket, 'inferred_pop.ht')
    #     if not can_reuse(
    #         [eigenvalues_ht_path, scores_ht_path, loadings_ht_path, inferred_pop_ht_path],
    #         overwrite,
    #     ):
    #         pca_job = add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/ancestry_pca.py '
    #             f'--mt-for-pca {mt_for_pca_union_hgdp_path} '
    #             f'--meta-tsv {samples_tsv_path} '
    #             f'--min-pop-prob {cutoffs_d["pca"]["min_pop_prob"]} '
    #             f'--n-pcs {num_ancestry_pcs} '
    #             f'--out-eigenvalues-ht {eigenvalues_ht_path} '
    #             f'--out-scores-ht {scores_ht_path} '
    #             f'--out-loadings-ht {loadings_ht_path} '
    #             f'--out-inferred-pop-ht {inferred_pop_ht_path} '
    #             + (
    #                 f'--related-samples-to-drop-ht '
    #                 f'{intermediate_related_samples_to_drop_ht_path} '
    #                 if intermediate_related_samples_to_drop_ht_path
    #                 else ''
    #             )
    #             + f'--tmp-bucket {join(tmp_bucket, "ancestry")} '
    #             + (f'--overwrite ' if overwrite else '')
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             is_test=is_test,
    #             phantomjs=True,
    #             highmem=highmem_workers,
    #             depends_on=[flag_related_job, subset_for_pca_job],
    #         )
    #     else:
    #         pca_job = b.new_job(f'{job_name} [reuse]')
    #
    #     regressed_metrics_ht_path = join(sample_qc_bucket, 'regressed_metrics.ht')
    #     job_name = 'Sample QC regressed filters'
    #     if not can_reuse([regressed_metrics_ht_path], overwrite):
    #         regressed_filters_job = add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/sample_qc_regressed_filters.py '
    #             f'--pca-scores-ht {scores_ht_path} '
    #             f'--hail-sample-qc-ht {hail_sample_qc_ht_path} '
    #             f'--out-regressed-metrics-ht {regressed_metrics_ht_path} '
    #             f'--tmp-bucket {tmp_bucket} '
    #             + (f'--overwrite ' if overwrite else '')
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             is_test=is_test,
    #             highmem=highmem_workers,
    #             depends_on=[pca_job, sample_qc_hardfilter_job],
    #         )
    #     else:
    #         regressed_filters_job = b.new_job(f'{job_name} [reuse]')
    #
    #     job_name = 'Plot PCA and loadings'
    #     out_path_ptn = join(ancestry_web_bucket, '{scope}_pc{pci}.{ext}')
    #     paths = []
    #     for scope in ['project', 'population', 'loadings']:
    #         for ext in ['html']:
    #             paths.append(
    #                 out_path_ptn.format(scope=scope, pci=num_ancestry_pcs - 1, ext=ext)
    #             )
    #     if not can_reuse(paths, overwrite):
    #         add_job(
    #             b,
    #             f'{join(utils.SCRIPTS_DIR, "ancestry_plot.py")} '
    #             f'--meta-tsv {samples_tsv_path} '
    #             f'--eigenvalues-ht {eigenvalues_ht_path} '
    #             f'--scores-ht {scores_ht_path} '
    #             f'--loadings-ht {loadings_ht_path} '
    #             f'--inferred-pop-ht {inferred_pop_ht_path} '
    #             + f'--out-path-pattern {out_path_ptn} '
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             is_test=is_test,
    #             phantomjs=True,
    #             highmem=highmem_workers,
    #             depends_on=[pca_job, regressed_filters_job],
    #         )
    #     else:
    #         b.new_job(f'{job_name} [reuse]')
    #
    #     job_name = f'Write sample QC metadata'
    #     meta_ht_path = join(analysis_bucket, 'meta.ht')
    #     meta_tsv_path = join(analysis_bucket, 'meta.tsv')
    #     if not can_reuse(
    #         [
    #             meta_ht_path,
    #             meta_tsv_path,
    #         ],
    #         overwrite,
    #     ):
    #         metadata_qc_job = add_job(
    #             b,
    #             f'{utils.SCRIPTS_DIR}/sample_qc_write_metadata.py '
    #             f'--meta-tsv {samples_tsv_path} '
    #             f'--hard-filtered-samples-ht {hard_filtered_samples_ht_path} '
    #             f'--sex-ht {sex_ht_path} '
    #             f'--hail-sample-qc-ht {hail_sample_qc_ht_path} '
    #             f'--regressed-filtes-ht {regressed_metrics_ht_path} '
    #             f'--relatedness-ht {relatedness_ht_path} '
    #             + (f'--release-related ' if release_related else '')
    #             + f'--pop-ht {inferred_pop_ht_path} '
    #             f'--tmp-bucket {tmp_bucket} '
    #             f'--out-meta-ht {meta_ht_path} '
    #             f'--out-meta-tsv {meta_tsv_path} '
    #             + (f'--overwrite ' if overwrite else '')
    #             + f'{_make_filter_cutoff_param(filter_cutoffs_path, sample_qc_bucket)} '
    #             + (f'--hail-billing {billing_project} ' if billing_project else ''),
    #             job_name=job_name,
    #             num_workers=scatter_count,
    #             is_test=is_test,
    #             highmem=highmem_workers,
    #             depends_on=[regressed_filters_job, pca_job, sample_qc_hardfilter_job],
    #         )
    #     else:
    #         metadata_qc_job = b.new_job(f'{job_name} [reuse]')
    #
    #     return metadata_qc_job, hard_filtered_samples_ht_path, meta_ht_path


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Generate final annotated, soft-filtered Matrix Table
"""

import logging
from typing import Optional

import click
import hail as hl

from joint_calling.utils import get_validation_callback, file_exists
from joint_calling import utils, _version

logger = logging.getLogger('joint-calling')
logger.setLevel('INFO')


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--mt',
    'mt_path',
    required=True,
    callback=get_validation_callback(ext='mt', must_exist=True),
    help='path to the raw sparse Matrix Table generated by combine_gvcfs.py',
)
@click.option(
    '--out-mt',
    'out_mt_path',
    required=True,
    callback=get_validation_callback(ext='mt', must_exist=False),
    help='path to write the final annotated soft-filtered Matrix Table',
)
@click.option(
    '--out-nonref-mt',
    'out_nonref_mt_path',
    callback=get_validation_callback(ext='mt', must_exist=False),
    help='write a version of output Matrix Table without reference blocks',
)
@click.option(
    '--meta-ht',
    'meta_ht_path',
    required=True,
    help='Table generated by sample_qc.py',
)
@click.option(
    '--freq-ht',
    'freq_ht_path',
    required=True,
    help='Table with frequency annotations',
)
@click.option(
    '--vqsr-ht',
    'vqsr_ht_path',
    required=True,
    help='Table with AS-VQSR annotations (both input and output)',
)
@click.option(
    '--allele-data-ht',
    'allele_data_ht_path',
    required=True,
    help='Table with allele_data annotation',
)
@click.option(
    '--qc-ac-ht',
    'qc_ac_ht_path',
    required=True,
    help='Table with qc_ac_* annotations',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    help='Hail billing account ID.',
)
def main(
    mt_path: str,
    out_mt_path: str,
    out_nonref_mt_path: Optional[str],
    meta_ht_path: str,
    vqsr_ht_path: str,
    freq_ht_path: str,
    allele_data_ht_path: str,
    qc_ac_ht_path: str,
    local_tmp_dir: str,
    overwrite: bool,  # pylint: disable=unused-argument
    hail_billing: str,  # pylint: disable=unused-argument
):  # pylint: disable=missing-function-docstring
    utils.init_hail('make_finalised_mt', local_tmp_dir)

    if file_exists(out_mt_path):
        if overwrite:
            logger.info(f'Output {out_mt_path} exists and will be overwritten')
        else:
            logger.error(
                f'Output file {out_mt_path} exists, use --overwrite to overwrite'
            )
            return

    # This will return a table with split multiallelics, and thus GT as a field
    mt = utils.get_mt(
        mt_path, split=True, add_meta=True, meta_ht=hl.read_table(meta_ht_path)
    )

    vqsr_ht = hl.read_table(vqsr_ht_path)
    mt = annotate_vqsr(mt, vqsr_ht)
    # INFO/AS_* fields
    # AS-VQSR output fields (AS_VQSLOD, AS_culprit, NEGATIVE_TRAIN_SITE, POSITIVE_TRAIN_SITE, AS_FilterStatus)
    # filters (corresponds to VCF's FILTER column. Populated from AS_FilterStatus after splitting by allele)
    
    freq_ht = hl.read_table(freq_ht_path)
    mt = mt.annotate_rows(**freq_ht[mt.row_key])
    # InbreedingCoeff
    # freq
    # faf
    # popmax:
    #   AC
    #   AF
    #   AN
    #   homozygote_count
    #   pop
    #   faf95
    # qual_hists
    # raw_qual_hists
    
    allele_data_ht = hl.read_table(allele_data_ht_path)
    mt = mt.annotate_rows(**allele_data_ht[mt.row_key])
    # allele_data:
    #   nonsplit_alleles
    #   has_star
    #   variant_type
    #   n_alt_alleles
    #   allele_type
    #   was_mixed
    
    qc_ac_ht = hl.read_table(qc_ac_ht_path)
    mt = mt.annotate_rows(**qc_ac_ht[mt.row_key])
    # ac_qc_samples_raw
    # ac_qc_samples_unrelated_raw
    # ac_release_samples_raw
    # ac_qc_samples_adj
    # ac_qc_samples_unrelated_adj
    # ac_release_samples_adj

    for ht in [vqsr_ht, freq_ht, allele_data_ht, qc_ac_ht]:
        mt = mt.annotate_globals(**ht.index_globals())

    mt.write(out_mt_path, overwrite=True)

    if out_nonref_mt_path:
        # hl.len(mt.alleles) > 1 - find sites with an ALT allele
        # hl.agg.any(mt.GT.is_non_ref() - check that there is any sample with non-ref
        #    genotype present, in case if all samples with ALT were filtered out
        mt = mt.filter_rows((hl.len(mt.alleles) > 1) & hl.agg.any(mt.GT.is_non_ref()))
        mt.write(out_nonref_mt_path, overwrite=True)


def annotate_vqsr(mt, vqsr_ht):
    """
    Assuming `vqsr_ht` is a site-only table (split multiallelics) from VQSR VCF, 
    and `mt` is a matrix table (split multiallelics), annotates `mt` rows from `vqsr_ht`
    """
    mt = mt.annotate_rows(**vqsr_ht[mt.row_key])
    
    # vqsr_ht has info annotation split by allele; plus new AS-VQSR annotations
    mt = mt.annotate_rows(info=vqsr_ht[mt.row_key].info)

    # populating filters which is outside of info
    mt = mt.annotate_rows(
        filters=mt.filters.union(vqsr_ht[mt.row_key].filters),
    )
    
    mt = mt.annotate_globals(**vqsr_ht.index_globals())
    return mt


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python3

"""
This script parses "somalier relate" (https://github.com/brentp/somalier) outputs,
and returns a non-zero code if either sex or pedigree mismatches the data in a 
provided PED file.
"""

import contextlib
import logging
import subprocess
import sys
import tempfile
from os.path import join, basename
from typing import Optional

import pandas as pd
import click
from peddy import Ped

logger = logging.getLogger('check-pedigree')
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.option(
    '--somalier-samples',
    'somalier_samples_fpath',
    required=True,
    help='Path to somalier {prefix}.samples.tsv output file',
)
@click.option(
    '--somalier-pairs',
    'somalier_pairs_fpath',
    required=True,
    help='Path to somalier {prefix}.pairs.tsv output file',
)
@click.option(
    '--somalier-html',
    'somalier_html_fpath',
    help='Path to somalier {prefix}.html output file',
)
def main(
    somalier_samples_fpath: str,
    somalier_pairs_fpath: str,
    somalier_html_fpath: Optional[str],
):  # pylint: disable=missing-function-docstring
    if somalier_samples_fpath.startswith('gs://'):
        local_tmp_dir = tempfile.mkdtemp()
        cmd = f'gsutil cp {somalier_samples_fpath} {somalier_pairs_fpath}'
        subprocess.run(cmd, check=False, shell=True)
        local_somalier_pairs_fpath = join(local_tmp_dir, basename(somalier_pairs_fpath))
        local_somalier_samples_fpath = join(
            local_tmp_dir, basename(somalier_samples_fpath)
        )
    else:
        local_somalier_pairs_fpath = somalier_pairs_fpath
        local_somalier_samples_fpath = somalier_samples_fpath

    df = pd.read_csv(local_somalier_samples_fpath, delimiter='\t')
    df.sex = df.sex.apply(lambda x: {1: 'male', 2: 'female'}.get(x, 'unknown'))
    df.original_pedigree_sex = df.original_pedigree_sex.apply(
        lambda x: {'-9': 'unknown'}.get(x, x)
    )
    bad_samples = list(df[df.gt_depth_mean == 0.0].sample_id)
    if bad_samples:
        logger.info(
            f'Excluding samples with non enough coverage to make inference: '
            f'{", ".join(bad_samples)}'
        )
    logger.info('-' * 10)

    logger.info('* Checking sex *')
    missing_inferred_sex = df.sex == 'unknown'
    missing_provided_sex = df.original_pedigree_sex == 'unknown'
    mismatching_female = (df.sex == 'female') & (df.original_pedigree_sex == 'male')
    mismatching_male = (df.sex == 'male') & (df.original_pedigree_sex == 'female')
    mismatching_sex = mismatching_female | mismatching_male

    def _print_stats(df_filter):
        for _, row in df[df_filter].iterrows():
            logger.info(
                f'\t{row.sample_id} ('
                f'provided: {row.original_pedigree_sex}, '
                f'inferred: {row.sex}, '
                f'mean depth: {row.gt_depth_mean})'
            )

    if mismatching_sex.any():
        logger.info(f'Found PED samples with mismatching sex:')
        _print_stats(mismatching_sex)
    if missing_inferred_sex.any():
        logger.info(f'Samples with missing inferred sex:')
        _print_stats(missing_inferred_sex)
    if missing_provided_sex.any():
        logger.info(f'Samples with missing provided sex:')
        _print_stats(missing_provided_sex)

    if not mismatching_sex.any():
        if (missing_inferred_sex | missing_provided_sex).any():
            logger.info(
                f'Inferred sex and pedigree matches or can be inferred for all samples.'
            )
        else:
            logger.info(f'Inferred sex and pedigree matches for all samples.')
    logger.info('-' * 10)

    logger.info('* Checking relatedness *')
    ped = Ped(local_somalier_samples_fpath)
    sample_by_id = {s.sample_id: s for s in ped.samples()}

    pairs_provided_as_unrelated_but_inferred_related = []
    other_mismatching_pairs = []
    pairs_df = pd.read_csv(local_somalier_pairs_fpath, delimiter='\t')
    for idx, row in pairs_df.iterrows():
        s1 = row['#sample_a']
        s2 = row['sample_b']
        if s1 in bad_samples or s2 in bad_samples:
            continue

        inferred_rel = infer_relationship(row['relatedness'], row['ibs0'], row['ibs2'])
        # Supressing all logging output from peddy as it would clutter the logs
        with contextlib.redirect_stderr(None), contextlib.redirect_stdout(None):
            peddy_rel = ped.relation(sample_by_id[s1], sample_by_id[s2])

        def _match_peddy_with_inferred(peddy_rel):
            return {
                'unrelated': 'unrelated',
                'related at unknown level': 'unrelated',
                'mom-dad': 'unrelated',
                'parent-child': 'parent-child',
                'grandchild': 'below_first_degree',
                'niece/nephew': 'below_first_degree',
                'great-grandchild': 'below_first_degree',
                'cousins': 'below_first_degree',
                'full siblings': 'siblings',
                'siblings': 'siblings',
                'unknown': 'unknown',
            }.get(peddy_rel)

        if (
            _match_peddy_with_inferred(peddy_rel) == 'unknown'
            and inferred_rel != 'unknown'
            or _match_peddy_with_inferred(peddy_rel) == 'unrelated'
            and inferred_rel != 'unrelated'
        ):
            pairs_provided_as_unrelated_but_inferred_related.append(
                f'"{s1}" and "{s2}", '
                f'provided: "{peddy_rel}", '
                f'inferred: "{inferred_rel}" '
                f'(rel={row["relatedness"]})'
            )
        elif inferred_rel != _match_peddy_with_inferred(peddy_rel):
            other_mismatching_pairs.append(
                f'"{s1}" and "{s2}", '
                f'provided: "{peddy_rel}", '
                f'inferred: "{inferred_rel}" '
                f'(rel={row["relatedness"]})'
            )
        pairs_df.loc[idx, 'provided_rel'] = peddy_rel
        pairs_df.loc[idx, 'inferred_rel'] = inferred_rel

    if pairs_provided_as_unrelated_but_inferred_related:
        logger.info(
            f'Found sample pairs that are provided as unrelated, but '
            f'inferred as related:'
        )
        for pair in pairs_provided_as_unrelated_but_inferred_related:
            logger.info(f'\t{pair}')
    if other_mismatching_pairs:
        logger.info(f'Found sample pairs with mismatched relatedness:')
        for pair in other_mismatching_pairs:
            logger.info(f'\t{pair}')
    else:
        logger.info(f'Inferred pedigree matches for all samples.')

    logger.info('-' * 10)
    print_info(
        df,
        pairs_df,
        somalier_samples_fpath,
        somalier_pairs_fpath,
        somalier_html_fpath,
    )
    if mismatching_sex.any() or other_mismatching_pairs:
        sys.exit(1)


def infer_relationship(coeff: float, ibs0: float, ibs2: float) -> str:
    """
    Inferres relashionship labels based on the kin coefficient
    and ibs0 and ibs2 values.
    """
    if coeff < 0.2:
        result = 'unrelated'
    elif coeff < 0.38:
        result = 'below_first_degree'
    elif coeff <= 0.62:
        if ibs0 / ibs2 < 0.005:
            result = 'parent-child'
        elif 0.015 < ibs0 / ibs2 < 0.052:
            result = 'siblings'
        else:
            result = 'first_degree'
    elif coeff < 0.8:
        result = 'first_degree_or_duplicate_or_twins'
    elif coeff >= 0.8:
        result = 'duplicate_or_twins'
    else:
        result = 'nan'
    return result


def print_info(
    samples_df,
    pairs_df,
    somalier_samples_fpath,
    somalier_pairs_fpath,
    somalier_html_fpath,
):
    """
    Print useful information to manually review pedigree check results
    """
    if len(samples_df) < 400:
        samples_str = samples_df.to_string()
        logger.info('')
        logger.info(
            f'Somalier results, samples (based on {somalier_samples_fpath}):\n{samples_str}\n'
        )
    if len(pairs_df) < 400:
        pairs_str = pairs_df[
            [
                '#sample_a',
                'sample_b',
                'relatedness',
                'ibs0',
                'ibs2',
                'n',
                'expected_relatedness',
            ]
        ].to_string()
        logger.info(
            f'Somalier results, sample pairs (based on {somalier_pairs_fpath}):\n{pairs_str}\n'
        )
    if somalier_html_fpath:
        logger.info(f'Somalier HTML report: {somalier_html_fpath}\n')


if __name__ == '__main__':
    main()  # pylint: disable=E1120



import importlib.metadata
from typing import Callable

import coloredlogs
import logging
import collections
from functools import lru_cache

from cpg_utils.config import get_config
from cpg_utils import Path, to_path
from cpg_utils.hail_batch import dataset_path

from hailtop.batch import Batch
from hailtop.batch.job import Job

from larcoh.dataproc import add_job
from larcoh.hb.batch import setup_batch, RegisteringBatch
from larcoh.utils import can_reuse, timestamp, slugify, exists
from larcoh.filetypes import SequencingType
from larcoh.inputs.metamist.inputs import CpgInputProvider
from larcoh.inputs.metamist.metamist import Metamist
from larcoh.targets import Cohort


logger = logging.getLogger(__file__)


coloredlogs.install(
    level='DEBUG', fmt='%(asctime)s %(levelname)s (%(name)s %(lineno)s): %(message)s'
)


def get_package_name() -> str:
    """
    Get name of the package.
    """
    return __name__.split('.', 1)[0]


def get_package_path() -> Path:
    """
    Get local install path of the package.
    """
    return to_path(__file__).parent.absolute()


def get_version() -> str:
    """
    Get package version.
    """
    return importlib.metadata.version(get_package_name())


output_version = get_config()['workflow']['output_version']

analysis_prefix = to_path(
    dataset_path(f'joint-calling/{output_version}', category='analysis')
)

tmp_prefix = to_path(dataset_path(f'joint-calling/{output_version}', category='tmp'))


@lru_cache
def cohort() -> Cohort:
    """
    Read inputs from the database.
    """
    c = Cohort(
        analysis_dataset_name=get_config()['workflow']['dataset'],
        sequencing_type=SequencingType.parse(
            get_config()['workflow']['sequencing_type']
        ),
    )

    CpgInputProvider(db=Metamist(c.analysis_dataset.name)).populate_cohort(
        cohort=c,
        dataset_names=get_config()['workflow'].get('datasets'),
        skip_samples=get_config()['workflow'].get('skip_samples'),
        only_samples=get_config()['workflow'].get('only_samples'),
        skip_datasets=get_config()['workflow'].get('skip_datasets'),
    )

    if get_config()['workflow'].get('access_level') == 'test':
        # Function that subsets the cohort for test runs
        original_sample_cnt = len(c.get_samples())
        original_dataset_cnt = len(c.get_datasets())
        # Test dataset: collecting all batch1 samples, plus 4 samples from each of other batches
        samples_by_batch: dict[str, list] = collections.defaultdict(list)
        for sample in c.get_samples():
            batch_id = sample.sequencing_meta_by_type[SequencingType.GENOME]['batch']
            if batch != '1' and len(samples_by_batch[batch_id]) > 4:
                sample.active = False
                continue
            samples_by_batch[batch_id].append(sample)
        logger.info(
            f'After subset: {len(c.get_samples())}/{original_sample_cnt} samples, '
            f'{len(samples_by_batch)} batches, '
            f'{len(c.get_datasets())}/{original_dataset_cnt} datasets'
        )
        logger.info(
            f'Using {len(c.get_samples())} samples '
            f'from {len(c.get_datasets())} datasets'
        )

    return c


@lru_cache
def batch() -> RegisteringBatch:
    """
    Initialise Batch object
    """
    run_id = get_config()['workflow'].get('run_id', timestamp())
    analysis_dataset = get_config()['workflow']['dataset']
    name = get_config()['workflow'].get('name')
    description = get_config()['workflow'].get('description')
    name = name or description or analysis_dataset
    name = slugify(name)
    description = description or name
    description += f': run_id={run_id}'

    description += f' [{cohort().sequencing_type.value}]'
    if ds_set := set(d.name for d in cohort().get_datasets()):
        description += ' ' + ', '.join(sorted(ds_set))
    return setup_batch(description=description)


def dataproc_job(
    script_name: str,
    params: dict,
    preemptible: bool = True,
    phantomjs: bool = True,
    num_workers: int | None = None,
    depends_on: list[Job] = None,
    autoscaling_policy: str | None = None,
    long: bool = False,
) -> Job:
    param_line = ' '.join(
        f'--{k.replace("_", "-")} {v}' for k, v in params.items() if v is not None
    )
    if num_workers is None:
        num_workers = get_config()['workflow']['scatter_count']

    return add_job(
        batch(),
        f'scripts/{script_name} {param_line}',
        job_name=script_name,
        num_workers=num_workers,
        highmem=get_config()['workflow'].get('highmem_workers'),
        preemptible=preemptible,
        phantomjs=phantomjs,
        depends_on=depends_on or [],
        autoscaling_policy=autoscaling_policy,
        long=long,
    )


# def queue_stage(
#     func: Callable[[...], list[Job]],
#     name: str,
#     output_paths: list[Path] | None = None,
#     **kwargs,
# ) -> list[Job]:
#     """
#     Queue jobs for one stage, checking inputs and outputs.
#     """
#     if output_paths and not can_reuse(output_paths):
#         return [batch().new_job(f'{name} [reuse]')]
#
#     return func(*output_paths, **kwargs)



import logging
import pickle

import click
import hail as hl

from gnomad.utils.vcf import (
    build_vcf_export_reference,
    rekey_new_reference,
)
from joint_calling import utils, _version
from joint_calling.utils import get_validation_callback


logger = logging.getLogger(__file__)
logger.setLevel('INFO')


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--ht',
    'ht_path',
    required=True,
    callback=get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--vcf-header-txt',
    'vcf_header_txt_path',
    required=True,
    callback=get_validation_callback(ext='txt', must_exist=True),
)
@click.option(
    '--out-vcf',
    'out_vcf_path',
    required=True,
    callback=get_validation_callback(ext='vcf.bgz'),
)
@click.option(
    '--name',
    'name',
    required=True,
    help='name of the dataset',
)
@click.option(
    '--chromosome',
    'chromosome',
    help='write a VCF for a chromosome',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
def main(
    ht_path: str,
    vcf_header_txt_path: str,
    out_vcf_path: str,
    name: str,
    chromosome: str,
    local_tmp_dir: str,
):  # pylint: disable=missing-function-docstring
    utils.init_hail(__file__, local_tmp_dir)

    ht = hl.read_table(ht_path)

    logger.info('Loading VCF header dict...')
    with hl.hadoop_open(vcf_header_txt_path, 'rb') as f:
        header_dict = pickle.load(f)

    if chromosome:
        logger.info(f'Exporting chromosome {chromosome}....')
        ht = hl.filter_intervals(ht, [hl.parse_locus_interval(chromosome)])

    export_reference = build_vcf_export_reference(name)

    hl.export_vcf(
        rekey_new_reference(ht, export_reference),
        out_vcf_path,
        metadata=header_dict,
        tabix=True,
    )


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Wrappers for bioinformatics file types (CRAM, GVCF, FASTQ, etc).
"""

from enum import Enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union, Optional

from hailtop.batch import ResourceGroup, ResourceFile, Batch

from cpg_utils import Path, to_path

from larcoh import utils


class SequencingType(Enum):
    """
    Type (scope) of a sequencing experiment.
    """

    GENOME = 'genome'
    EXOME = 'exome'
    SINGLE_CELL = 'single_cell'
    MTSEQ = 'mtseq'
    ONT = 'ont'

    def seqr_value(self) -> str:
        """
        Map to Seqr-style string
        https://github.com/broadinstitute/seqr/blob/e0c179c36c0f68c892017de5eab2e4c1b9ffdc92/seqr/models.py#L592-L594
        """
        return {
            SequencingType.GENOME: 'WGS',
            SequencingType.EXOME: 'WES',
            SequencingType.SINGLE_CELL: 'RNA',
        }.get(self, '')

    @staticmethod
    def parse(str_val: str) -> 'SequencingType':
        """
        Parse a string into a SequencingType object.

        >>> SequencingType.parse('genome')
        SequencingType.GENOME
        >>> SequencingType.parse('wes')
        SequencingType.EXOME
        """
        str_to_val: dict[str, SequencingType] = {}
        for val, str_vals in {
            SequencingType.GENOME: ['genome', 'wgs'],
            SequencingType.EXOME: ['exome', 'wts', 'wes'],
            SequencingType.SINGLE_CELL: ['single_cell', 'single_cell_rna'],
            SequencingType.MTSEQ: ['mtseq'],
        }.items():
            for str_v in str_vals:
                str_v = str_v.lower()
                assert str_v not in str_to_val, (str_v, str_to_val)
                str_to_val[str_v] = val
        str_val = str_val.lower()
        if str_val not in str_to_val:
            raise ValueError(
                f'Unrecognised sequence type {str_val}. '
                f'Available: {list(str_to_val.keys())}'
            )
        return str_to_val[str_val]


class AlignmentInput(ABC):
    """
    Data that works as input for alignment or realignment.
    """

    @abstractmethod
    def exists(self) -> bool:
        """
        Check that all files exist.
        """

    @abstractmethod
    def path_glob(self) -> str:
        """
        Compact representation of file paths.
        """


class CramPath(AlignmentInput):
    """
    Represents alignment data on a bucket within the pipeline.
    Includes a path to a CRAM or a BAM file along with a corresponding index,
    and a corresponding fingerprint path.
    """

    def __init__(
        self,
        path: str | Path,
        reference_assembly: str = None,
        index_path: Path | str | None = None,
    ):
        self.path = to_path(path)
        self.is_bam = self.path.suffix == '.bam'
        self.ext = 'cram' if not self.is_bam else 'bam'
        self.index_ext = 'crai' if not self.is_bam else 'bai'
        self._index_path = index_path
        self.somalier_path = to_path(f'{self.path}.somalier')
        self.sequencing_type: Optional[SequencingType] = None
        self.reference_assembly = reference_assembly

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self) -> str:
        return f'CRAM({self.path})'

    def exists(self) -> bool:
        """
        CRAM file exists.
        """
        return utils.exists(self.path)

    def index_exists(self) -> bool:
        """
        CRAI/BAI index exists
        """
        return utils.exists(self.index_path)

    @property
    def index_path(self) -> Path:
        """
        Path to the corresponding CRAI/BAI index
        """
        return (
            to_path(self._index_path)
            if self._index_path
            else to_path(f'{self.path}.{self.index_ext}')
        )

    def resource_group(self, b: Batch) -> ResourceGroup:
        """
        Create a Hail Batch resource group
        """
        d = {
            self.ext: str(self.path),
        }
        if self.index_exists():
            d[f'{self.ext}.{self.index_ext}'] = str(self.index_path)

        return b.read_input_group(**d)

    def path_glob(self) -> str:
        """
        Compact representation of file paths.
        For a CRAM, it's just the CRAM file path.
        """
        return str(self.path)


class GvcfPath:
    """
    Represents GVCF data on a bucket within the pipeline.
    Includes a path to a GVCF file along with a corresponding TBI index,
    and a corresponding fingerprint path.
    """

    def __init__(self, path: Path | str):
        self.path = to_path(path)
        self.somalier_path = to_path(f'{self.path}.somalier')

    def __str__(self) -> str:
        return str(self.path)

    def __repr__(self) -> str:
        return f'GVCF({self.path})'

    def exists(self) -> bool:
        """
        GVCF file exists.
        """
        return utils.exists(self.path)

    @property
    def tbi_path(self) -> Path:
        """
        Path to the corresponding index
        """
        return to_path(f'{self.path}.tbi')

    def resource_group(self, b: Batch) -> ResourceGroup:
        """
        Create a Hail Batch resource group
        """
        return b.read_input_group(
            **{
                'g.vcf.gz': str(self.path),
                'g.vcf.gz.tbi': str(self.tbi_path),
            }
        )


FastqPath = Union[str, Path, ResourceFile]


@dataclass
class FastqPair:
    """
    Pair of FASTQ files
    """

    r1: FastqPath
    r2: FastqPath

    def __getitem__(self, i):
        assert i == 0 or i == 1, i
        return [self.r1, self.r2][i]

    def as_resources(self, b) -> 'FastqPair':
        """
        Makes a pair of ResourceFile objects for r1 and r2.
        """
        return FastqPair(
            *[
                self[i]
                if isinstance(self[i], ResourceFile)
                else b.read_input(str(self[i]))
                for i in [0, 1]
            ]
        )

    def __str__(self):
        return f'{self.r1}|{self.r2}'


class FastqPairs(list[FastqPair], AlignmentInput):
    """
    Multiple FASTQ file pairs belonging to the same sample
    (e.g. multiple lanes or top-ups).
    """

    def exists(self) -> bool:
        """
        Check if each FASTQ file in each pair exist.
        """
        return all(utils.exists(pair.r1) and utils.exists(pair.r2) for pair in self)

    def path_glob(self) -> str:
        """
        Compact representation of file paths.
        For FASTQ pairs, it's glob string to find all FASTQ files.

        >>> FastqPairs([
        >>>     FastqPair('gs://sample_R1.fq.gz', 'gs://sample_R2.fq.gz'),
        >>> ]).path_glob()
        'gs://sample_R{2,1}.fq.gz'
        >>> FastqPairs([
        >>>     FastqPair('gs://sample_L1_R1.fq.gz', 'gs://sample_L1_R2.fq.gz'),
        >>>     FastqPair('gs://sample_L2_R1.fq.gz', 'gs://sample_L2_R2.fq.gz'),
        >>> ]).path_glob()
        'gs://sample_L{2,1}_R{2,1}.fq.gz'
        """
        all_fastq_paths = []
        for pair in self:
            all_fastq_paths.extend([pair.r1, pair.r2])
        # Triple braces are intentional: they are resolved to single ones.
        return ''.join(
            [
                f'{{{",".join(set(chars))}}}' if len(set(chars)) > 1 else chars[0]
                for chars in zip(*map(str, all_fastq_paths))
            ]
        )



#!/usr/bin/env python

"""
Remove ref blocks
"""

import logging

import click
import hail as hl

from joint_calling.utils import get_validation_callback, file_exists
from joint_calling import utils, _version

logger = logging.getLogger('joint-calling')
logger.setLevel('INFO')


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--mt',
    'mt_path',
    required=True,
    callback=get_validation_callback(ext='mt', must_exist=True),
    help='path to the raw sparse Matrix Table generated by combine_gvcfs.py',
)
@click.option(
    '--out-mt',
    'out_mt_path',
    required=True,
    callback=get_validation_callback(ext='mt', must_exist=False),
    help='path to write the final annotated soft-filtered Matrix Table',
)
@click.option(
    '--local-tmp-dir',
    'local_tmp_dir',
    help='local directory for temporary files and Hail logs (must be local).',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    help='Hail billing account ID.',
)
def main(
    mt_path: str,
    out_mt_path: str,
    local_tmp_dir: str,
    overwrite: bool,  # pylint: disable=unused-argument
    hail_billing: str,  # pylint: disable=unused-argument
):  # pylint: disable=missing-function-docstring
    utils.init_hail('make_noref_mt', local_tmp_dir)

    if file_exists(out_mt_path):
        if overwrite:
            logger.info(f'Output {out_mt_path} exists and will be overwritten')
        else:
            logger.error(
                f'Output file {out_mt_path} exists, use --overwrite to overwrite'
            )
            return
    
    mt = hl.read_matrix_table(mt_path)
    
    # hl.len(mt.alleles) > 1 - find sites with an ALT allele
    # hl.agg.any(mt.GT.is_non_ref() - check that there is any sample with non-ref
    #    genotype present, in case if all samples with ALT were filtered out
    mt = mt.filter_rows((hl.len(mt.alleles) > 1) & hl.agg.any(mt.GT.is_non_ref()))
    mt.write(out_mt_path, overwrite=True)


if __name__ == '__main__':
    main()  # pylint: disable=E1120



#!/usr/bin/env python

"""
Run sample QC on a MatrixTable, hard filter samples, add soft filter labels.
"""

import logging
import pickle
from os.path import join
from typing import Optional, List, Dict

import click
import hail as hl
import pandas as pd
from cpg_utils.config import get_config
from cpg_utils.hail_batch import reference_path
from gnomad.sample_qc.ancestry import run_pca_with_relateds, assign_population_pcs
from gnomad.sample_qc.filtering import (
    compute_qc_metrics_residuals,
    compute_stratified_metrics_filter,
)
from gnomad.sample_qc.filtering import compute_stratified_sample_qc
from gnomad.sample_qc.pipeline import annotate_sex
from gnomad.sample_qc.relatedness import compute_related_samples_to_drop
from gnomad.utils.annotations import bi_allelic_expr
from gnomad.utils.file_utils import file_exists
from gnomad.utils.filtering import filter_to_autosomes
from gnomad.utils.sparse_mt import filter_ref_blocks

from larcoh import utils, can_reuse

logger = logging.getLogger(__file__)


@click.command()
@click.option(
    '--mt',
    'mt_path',
    required=True,
    callback=utils.get_validation_callback(ext='mt', must_exist=True),
    help='path to the input Matrix Table. '
    'To generate it, run the `combine_gvcfs` script',
)
@click.option(
    '--cohort-tsv',
    'cohort_tsv_path',
    required=True,
    help='path to a CSV with QC metadata for the samples in the input Matrix Table. '
    'The following columns are expected: '
    's,freemix,pct_chimeras,duplication,insert_size. '
    'Must be keyed by "s".',
)
@click.option(
    '--out-hard-filtered-samples-ht',
    'out_hard_filtered_samples_ht_path',
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
    required=True,
)
@click.option(
    '--out-sex-ht',
    'out_sex_ht_path',
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
    required=True,
)
@click.option(
    '--out-hail-sample-qc-ht',
    'out_hail_sample_qc_ht_path',
    callback=utils.get_validation_callback(ext='ht', must_exist=False),
    required=True,
)
@click.option(
    '--tmp-bucket',
    'tmp_bucket',
    required=True,
    help='path to write temporary intermediate files and checkpoints '
    '(usually a temporary bucket)',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    mt_path: str,
    cohort_tsv_path: str,
    out_hard_filtered_samples_ht_path: str,
    out_sex_ht_path: str,
    out_hail_sample_qc_ht_path: str,
    tmp_bucket: str,
):
    mt = utils.get_mt(mt_path, passing_sites_only=True)
    mt_split = utils.get_mt(mt_path, passing_sites_only=True, split=True)

    input_metadata_ht = hl.import_table(cohort_tsv_path)

    # `hail_sample_qc_ht` row fields: sample_qc, bi_allelic_sample_qc
    hail_sample_qc_ht = compute_hail_sample_qc(
        mt=mt_split,
        tmp_bucket=tmp_bucket,
        out_ht_path=out_hail_sample_qc_ht_path,
    )

    # `sex_ht` row fields: is_female, chr20_mean_dp, sex_karyotype
    sex_ht = infer_sex(
        mt=mt,
        tmp_bucket=tmp_bucket,
        out_ht_path=out_sex_ht_path,
    )

    # `out_hard_filtered_samples_ht_path` row fields: hard_filters;
    # also includes only failed samples
    compute_hard_filters(
        mt=mt,
        picard_metrics_ht=input_metadata_ht,
        sex_ht=sex_ht,
        hail_sample_qc_ht=hail_sample_qc_ht,
        out_ht_path=out_hard_filtered_samples_ht_path,
        cutoffs_d=get_config()['larhoc']['hardfiltering'],
    )


def compute_hail_sample_qc(
    mt: hl.MatrixTable,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
) -> hl.Table:
    """
    Runs Hail hl.sample_qc() to generate a Hail Table with
    per-sample QC metrics.
    :param mt: input MatrixTable, multiallelics split, genotype in GT fields
    :param tmp_bucket: bucket path to write checkpoints
    :param out_ht_path: location to write the result to
    :return: a Hail Table with the following row fields:
        'sample_qc': {
            <output of hl.sample_qc() (n_filtered, n_hom_ref, etc)>
        }
        'bi_allelic_sample_qc': {
            <output of hl.sample_qc() for bi-allelic variants only>
        }
        'multi_allelic_sample_qc': {
            <output of hl.sample_qc() for multi-allelic variants only>
        }
    """
    logger.info('Sample QC')
    out_ht_path = out_ht_path or join(tmp_bucket, 'hail_sample_qc.ht')
    if utils.can_reuse(out_ht_path):
        return hl.read_table(out_ht_path)

    mt = filter_to_autosomes(mt)

    mt = mt.select_entries('GT')

    mt = filter_ref_blocks(mt)

    # Remove centromeres and telomeres incase they were included and any reference blocks
    tel_cent_ht = hl.read_table(str(reference_path('gnomad/tel_and_cent_ht')))
    mt = mt.filter_rows(hl.is_missing(tel_cent_ht[mt.locus]))

    sample_qc_ht = compute_stratified_sample_qc(
        mt,
        strata={
            'bi_allelic': bi_allelic_expr(mt),
            'multi_allelic': ~bi_allelic_expr(mt),
        },
        tmp_ht_prefix=None,
    )

    # Remove annotations that cannot be computed from the sparse format
    sample_qc_ht = sample_qc_ht.annotate(
        **{
            x: sample_qc_ht[x].drop('n_called', 'n_not_called', 'call_rate')
            for x in sample_qc_ht.row_value
        }
    )

    sample_qc_ht = sample_qc_ht.repartition(100)
    return sample_qc_ht.checkpoint(out_ht_path, overwrite=True)


def infer_sex(
    mt: hl.MatrixTable,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
    target_regions: Optional[hl.Table] = None,
) -> hl.Table:
    """
    Runs gnomad.sample_qc.annotate_sex() to infer sex
    :param mt: input MatrixTable
    :param tmp_bucket: bucket to write checkpoints
    :param out_ht_path: location to write the result to
    :param target_regions: if exomes, it should correspond to the regions in the panel
    :return: a Table with the following row fields:
        'is_female': bool
        'f_stat': float64
        'n_called': int64
        'expected_homs': float64
        'observed_homs': int64
        'chr20_mean_dp': float32
        'chrX_mean_dp': float32
        'chrY_mean_dp': float32
        'chrX_ploidy': float32
        'chrY_ploidy': float32
        'X_karyotype': str
        'Y_karyotype': str
        'sex_karyotype': str
    """
    logger.info('Inferring sex')
    out_ht_path = out_ht_path or join(tmp_bucket, 'sex.ht')
    if utils.can_reuse(out_ht_path):
        return hl.read_table(out_ht_path)

    tel_cent_ht = hl.read_table(str(reference_path('gnomad/tel_and_cent_ht')))
    ht = annotate_sex(
        mt,
        excluded_intervals=tel_cent_ht,
        included_intervals=target_regions,
        gt_expr='LGT',
        normal_ploidy_cutoff=5,  # default value
        aneuploidy_cutoff=7,  # just a bit higher value
    )

    logger.info('Counting het/hom ratio on chrX as an extra sex check')
    # Number of hom/hem on chrX, as an extra sex check measure
    chrx_mt = hl.filter_intervals(mt, [hl.parse_locus_interval('chrX')])
    chrx_mt = chrx_mt.annotate_entries(GT=chrx_mt.LGT)
    chrx_ht = hl.sample_qc(chrx_mt).cols()
    ht = ht.annotate(chrX_sample_qc=chrx_ht[ht.key].sample_qc)
    ht = ht.annotate(chrX_r_het_hom_var=ht.chrX_sample_qc.r_het_hom_var)

    return ht.checkpoint(out_ht_path, overwrite=True)


def run_pca_ancestry_analysis(
    mt: hl.MatrixTable,
    sample_to_drop_ht: Optional[hl.Table],
    n_pcs: int,
    out_eigenvalues_ht_path: str,
    out_scores_ht_path: str,
    out_loadings_ht_path: str,
) -> hl.Table:
    """
    :param mt: variants usable for PCA analysis, combined with samples
        with known populations (HGDP, 1KG, etc)
    :param sample_to_drop_ht: table with samples to drop based on
        previous relatedness analysis. With a `rank` row field
    :param n_pcs: maximum number of principal components
    :param out_eigenvalues_ht_path: path to a txt file to write PCA eigenvalues
    :param out_scores_ht_path: path to write PCA scores
    :param out_loadings_ht_path: path to write PCA loadings
    :return: a Table with a row field:
        'scores': array<float64>
    """
    logger.info('Running PCA ancestry analysis')
    if all(
        not fp or utils.can_reuse(fp)
        for fp in [
            out_eigenvalues_ht_path,
            out_scores_ht_path,
            out_loadings_ht_path,
        ]
    ):
        return hl.read_table(out_scores_ht_path)

    # Adjusting the number of principal components not to exceed the
    # number of samples
    samples_to_drop_num = 0 if sample_to_drop_ht is None else sample_to_drop_ht.count()
    n_pcs = min(n_pcs, mt.cols().count() - samples_to_drop_num)

    logger.info(f'mt.s: {mt.s.collect()}')
    if sample_to_drop_ht:
        logger.info(f'sample_to_drop_ht.s: {sample_to_drop_ht.s.collect()}')

    eigenvalues, scores_ht, loadings_ht = run_pca_with_relateds(
        mt, sample_to_drop_ht, n_pcs=n_pcs
    )
    logger.info(f'scores_ht.s: {list(scores_ht.s.collect())}')
    logger.info(f'eigenvalues: {eigenvalues}')
    eigenvalues_ht = hl.Table.from_pandas(pd.DataFrame(eigenvalues, columns=['f0']))
    eigenvalues_ht.write(out_eigenvalues_ht_path, overwrite=True)
    loadings_ht.write(out_loadings_ht_path, overwrite=True)
    scores_ht.write(out_scores_ht_path, overwrite=True)
    scores_ht = hl.read_table(out_scores_ht_path)
    return scores_ht


def infer_pop_labels(
    scores_ht: hl.Table,
    training_pop_ht: hl.Table,
    tmp_bucket: str,
    min_prob: float,
    max_mislabeled_training_samples: int = 50,
    n_pcs: int = 16,
    out_ht_path: Optional[str] = None,
) -> hl.Table:
    """
    Take population PCA results and training data, and run random forest
    to assign global population labels.

    :param scores_ht: output table of `_run_pca_ancestry_analysis()`
        with a row field 'scores': array<float64>
    :param training_pop_ht: table with samples with defined `training_pop`: str
    :param tmp_bucket: bucket to write checkpoints and intermediate files
    :param min_prob: min probability of belonging to a given population
        for the population to be set (otherwise set to `None`)
    :param max_mislabeled_training_samples: keep rerunning until the number
        of mislabeled samples is below this number
    :param n_pcs: Number of PCs to use in the RF
    :param out_ht_path: Path to write the resulting HT table
    :return: a Table with the following row fields, including `prob_<POP>`
        probabily fields for each population label:
        'training_pop': str
        'pca_scores': array<float64>
        'pop': str
        'prob_CEU': float64
        'prob_YRI': float64
        ... (prob_*: float64 for each population label)
    """
    out_ht_path = out_ht_path or join(tmp_bucket, 'inferred_pop.ht')
    if utils.can_reuse(out_ht_path):
        return hl.read_table(out_ht_path)

    if training_pop_ht.count() < 2:
        logger.warning(
            'Need at least 2 samples with known `population` label to run PCA'
            'and assign population labels to remaining samples'
        )
        pop_ht = scores_ht.annotate(
            pop='oth',
            is_training=False,
            pca_scores=scores_ht.scores,
        )
        return pop_ht.checkpoint(out_ht_path, overwrite=True)

    logger.info(
        'Using calculated PCA scores as well as training samples with known '
        '`population` label to assign population labels to remaining samples'
    )
    scores_ht = scores_ht.annotate(
        training_pop=training_pop_ht[scores_ht.key].training_pop
    )

    def _run_assign_population_pcs(pop_pca_scores_ht, min_prob):
        examples_num = pop_pca_scores_ht.aggregate(
            hl.agg.count_where(hl.is_defined(pop_pca_scores_ht.training_pop))
        )
        logger.info(f'Running RF using {examples_num} training examples')
        pop_ht, pops_rf_model = assign_population_pcs(
            pop_pca_scores_ht,
            pc_cols=pop_pca_scores_ht.scores[:n_pcs],
            known_col='training_pop',
            min_prob=min_prob,
        )
        n_mislabeled_samples = pop_ht.aggregate(
            hl.agg.count_where(pop_ht.training_pop != pop_ht.pop)
        )
        return pop_ht, pops_rf_model, n_mislabeled_samples

    pop_ht, pops_rf_model, n_mislabeled_samples = _run_assign_population_pcs(
        scores_ht, min_prob
    )
    while n_mislabeled_samples > max_mislabeled_training_samples:
        logger.info(
            f'Found {n_mislabeled_samples} samples '
            f'labeled differently from their known pop. '
            f'Re-running without them.'
        )

        pop_ht = pop_ht[scores_ht.key]
        pop_pca_scores_ht = scores_ht.annotate(
            training_pop=hl.or_missing(
                (pop_ht.training_pop == pop_ht.pop), scores_ht.training_pop
            )
        ).persist()

        pop_ht, pops_rf_model, n_mislabeled_samples = _run_assign_population_pcs(
            pop_pca_scores_ht, min_prob
        )

    # Writing a tab delimited file indicating inferred sample populations
    pop_tsv_file = join(tmp_bucket, 'RF_pop_assignments.txt.gz')
    if not can_reuse(pop_tsv_file):
        pc_cnt = min(hl.min(10, hl.len(pop_ht.pca_scores)).collect())
        pop_ht.transmute(
            **{f'PC{i + 1}': pop_ht.pca_scores[i] for i in range(pc_cnt)}
        ).export(pop_tsv_file)

    # Writing the RF model used for inferring sample populations
    pop_rf_file = join(tmp_bucket, 'pop.RF_fit.pickle')
    if not can_reuse(pop_rf_file):
        with hl.hadoop_open(pop_rf_file, 'wb') as out:
            pickle.dump(pops_rf_model, out)

    pop_ht = pop_ht.annotate(is_training=hl.is_defined(training_pop_ht[pop_ht.key]))
    return pop_ht.checkpoint(out_ht_path, overwrite=True)


def compute_stratified_qc(
    sample_qc_ht: hl.Table,
    pop_ht: hl.Table,
    work_bucket: str,
    filtering_qc_metrics: List[str],
) -> hl.Table:
    """
    Computes median, MAD, and upper and lower thresholds for each metric
    in `filtering_qc_metrics`, groupped by `pop` field in `pop_ht`

    :param sample_qc_ht: table with a row field
        `sample_qc` = struct { n_snp: int64, n_singleton: int64, ... }
    :param pop_ht: table with a `pop` row field
    :param work_bucket: bucket to write checkpoints
    :param filtering_qc_metrics: metrics to annotate with
    :return: table with the following structure:
        Global fields:
            'qc_metrics_stats': dict<tuple (
                str
                ), struct {
                    n_snp: struct {
                        median: int64,
                        mad: float64,
                        lower: float64,
                        upper: float64
                    },
                    n_singleton: struct {
                        ...
                    },
                    ...
            }
        Row fields:
            'fail_n_snp': bool
            'fail_n_singleton': bool
            ...
            'qc_metrics_filters': set<str>
    """
    out_ht_path = join(work_bucket, 'stratified_metrics.ht')
    if can_reuse(out_ht_path):
        return hl.read_table(out_ht_path)

    logger.info(
        f'Computing stratified QC metrics filters using '
        f'metrics: {", ".join(filtering_qc_metrics)}'
    )

    sample_qc_ht = sample_qc_ht.annotate(qc_pop=pop_ht[sample_qc_ht.key].pop)
    stratified_metrics_ht = compute_stratified_metrics_filter(
        sample_qc_ht,
        qc_metrics={
            metric: sample_qc_ht.sample_qc[metric] for metric in filtering_qc_metrics
        },
        strata={'qc_pop': sample_qc_ht.qc_pop},
        metric_threshold={'n_singleton': (4.0, 8.0)},
    )

    stratified_metrics_ht.write(out_ht_path, overwrite=True)
    return hl.read_table(out_ht_path)


def flag_related_samples(
    hard_filtered_samples_ht: hl.Table,
    sex_ht: hl.Table,
    relatedness_ht: hl.Table,
    regressed_metrics_ht: Optional[hl.Table],
    tmp_bucket: str,
    kin_threshold: float,
    out_ht_path: Optional[str] = None,
) -> hl.Table:
    """
    Flag samples to drop based on relatedness, so the final set
    has only unrelated samples, best quality one per family

    :param hard_filtered_samples_ht: table with failed samples
        and a `hard_filters` row field
    :param sex_ht: table with a `chr20_mean_dp` row field
    :param relatedness_ht: table keyed by exactly two fields (i and j)
        of the same type, identifying the pair of samples for each row
    :param regressed_metrics_ht: optional table with a `qc_metrics_filters`
        field calculated with _apply_regressed_filters() from PCA scores
    :param tmp_bucket: bucket to write checkpoints
    :param kin_threshold: kinship threshold to call two samples as related
    :param out_ht_path: path to write the resulting Table to
    :return: a table of the samples to drop along with their rank
        row field: 'rank': int64
    """
    label = 'final' if regressed_metrics_ht is not None else 'intermediate'
    logger.info(f'Flagging related samples to drop, {label}')
    out_ht_path = out_ht_path or join(tmp_bucket, f'{label}_related_samples_to_drop.ht')
    if utils.can_reuse(out_ht_path):
        return hl.read_table(out_ht_path)

    rankings_ht_path = join(tmp_bucket, f'{label}_samples_rankings.ht')
    if can_reuse(rankings_ht_path):
        rank_ht = hl.read_table(rankings_ht_path)
    else:
        rank_ht = _compute_sample_rankings(
            hard_filtered_samples_ht,
            sex_ht,
            use_qc_metrics_filters=regressed_metrics_ht is not None,
            regressed_metrics_ht=regressed_metrics_ht,
        ).checkpoint(rankings_ht_path, overwrite=True)

    try:
        filtered_samples = hl.literal(
            rank_ht.aggregate(
                hl.agg.filter(rank_ht.filtered, hl.agg.collect(rank_ht.s))
            )
        )
    except hl.ExpressionException:
        # Hail doesn't handle it with `aggregate` when none of
        # the samples is 'filtered'
        filtered_samples = hl.empty_array('tstr')

    samples_to_drop_ht = compute_related_samples_to_drop(
        relatedness_ht,
        rank_ht,
        kin_threshold=kin_threshold,
        filtered_samples=filtered_samples,
    )
    return samples_to_drop_ht.checkpoint(out_ht_path, overwrite=True)


def _compute_sample_rankings(
    hard_filtered_samples_ht: hl.Table,
    sex_ht: hl.Table,
    use_qc_metrics_filters: bool = False,
    regressed_metrics_ht: Optional[hl.Table] = None,
) -> hl.Table:
    """
    Orders samples by hard filters and coverage and adds rank,
    which is the lower the better.

    :param hard_filtered_samples_ht: table with failed samples
        and a `hard_filters` row field
    :param sex_ht: table with a `chr20_mean_dp` row field
    :param use_qc_metrics_filters: apply population-stratified QC filters
    :param regressed_metrics_ht: table with a `qc_metrics_filters` field.
        Used only if `use_qc_metrics_filters` is True.
    :return: table ordered by rank, with the following row fields:
        `rank`, `filtered`
    """
    ht = sex_ht.drop(*list(sex_ht.globals.dtype.keys()))
    ht = ht.select(
        'chr20_mean_dp',
        filtered=hl.or_else(
            hl.len(hard_filtered_samples_ht[ht.key].hard_filters) > 0, False
        ),
    )
    if use_qc_metrics_filters and regressed_metrics_ht is not None:
        ht = ht.annotate(
            filtered=hl.cond(
                ht.filtered,
                True,
                hl.or_else(
                    hl.len(regressed_metrics_ht[ht.key].qc_metrics_filters) > 0, False
                ),
            )
        )

    ht = ht.order_by(ht.filtered, hl.desc(ht.chr20_mean_dp)).add_index(name='rank')
    return ht.key_by('s').select('filtered', 'rank')


def apply_regressed_filters(
    sample_qc_ht: hl.Table,
    pop_pca_scores_ht: hl.Table,
    tmp_bucket: str,
    out_ht_path: Optional[str] = None,
) -> hl.Table:
    """
    Re-compute QC metrics (with hl.sample_qc() - like n_snp, r_het_hom)
    per population, and adding "fail_*" row fields when a metric is below
    the the lower MAD threshold or higher the upper MAD threshold
    (see `compute_stratified_metrics_filter` for defaults)

    :param sample_qc_ht: table with a row field
       `bi_allelic_sample_qc` =
          struct { n_snp: int64, n_singleton: int64, ... }
    :param pop_pca_scores_ht: table with a `scores` row field
    :param tmp_bucket: bucket to write checkpoints
    :param out_ht_path: path to write the resulting table to
    :return: a table with the folliwing structure:
        Global fields:
            'lms': struct {
                n_snp: struct {
                    beta: array<float64>,
                    standard_error: array<float64>,
                    t_stat: array<float64>,
                    p_value: array<float64>,
                    multiple_standard_error: float64,
                    multiple_r_squared: float64,
                    adjusted_r_squared: float64,
                    f_stat: float64,
                    multiple_p_value: float64,
                    n: int32
                },
                n_singleton: struct {
                    ...
                },
                ...
            }
            'qc_metrics_stats': struct {
                n_snp_residual: struct {
                    median: float64,
                    mad: float64,
                    lower: float64,
                    upper: float64
                },
                n_singleton_residual: struct {
                    ...
                },
                ...
            }
        Row fields:
            's': str
            'n_snp_residual': float64
            'n_singleton_residual': float64
            ...
            'fail_n_snp_residual': bool
            'fail_n_singleton_residual': bool
            ...
            'qc_metrics_filters': set<str>
    """
    logger.info('Compute QC metrics adjusted for popopulation')
    out_ht_path = out_ht_path or join(tmp_bucket, 'regressed_metrics.ht')
    if utils.can_reuse(out_ht_path):
        return hl.read_table(out_ht_path)

    sample_qc_ht = sample_qc_ht.select(
        **sample_qc_ht.bi_allelic_sample_qc,
        **pop_pca_scores_ht[sample_qc_ht.key],
        releasable=hl.bool(True),
    )

    filtering_qc_metrics = [
        'n_snp',
        'n_singleton',
        'r_ti_tv',
        'r_insertion_deletion',
        'n_insertion',
        'n_deletion',
        'r_het_hom_var',
        'n_het',
        'n_hom_var',
        'n_transition',
        'n_transversion',
    ]
    residuals_ht = compute_qc_metrics_residuals(
        ht=sample_qc_ht,
        pc_scores=sample_qc_ht.scores,
        qc_metrics={metric: sample_qc_ht[metric] for metric in filtering_qc_metrics},
        regression_sample_inclusion_expr=sample_qc_ht.releasable,
    )

    stratified_metrics_ht = compute_stratified_metrics_filter(
        ht=residuals_ht,
        qc_metrics=dict(residuals_ht.row_value),
        metric_threshold={'n_singleton_residual': (4.0, 8.0)},
    )

    residuals_ht = residuals_ht.annotate(**stratified_metrics_ht[residuals_ht.key])
    residuals_ht = residuals_ht.annotate_globals(
        **stratified_metrics_ht.index_globals()
    )

    return residuals_ht.checkpoint(out_ht_path, overwrite=True)


def compute_hard_filters(
    mt: hl.MatrixTable,
    picard_metrics_ht: hl.Table,
    sex_ht: hl.Table,
    hail_sample_qc_ht: hl.Table,
    cutoffs_d: Dict,
    out_ht_path: str,
) -> hl.Table:
    """
    Uses the sex imputation results, results of the sample_qc() run on
    bi-allelic variants, and stats by Picard tools specificed in `sample_df`,
    to apply filters to samples in `mt`, and create a table with
    only samples that fail at least one sample.

    :param mt: input matrix table
    :param picard_metrics_ht: table QC metrics from Picard tools. Expected
        fields: r_contamination, r_chimera, r_duplication, insert_size
    :param sex_ht: required fields: "sex_karyotype", "chr20_mean_dp"
    :param hail_sample_qc_ht: required fields:
        "bi_allelic_sample_qc { n_snp, n_singleton, r_het_hom_var }"
    :param out_ht_path: location to write the hard filtered samples Table
    :param cutoffs_d: a dictionary with hard-filtering thresholds
    :return: a table with samples failed the filters, and the following structure:
        's': str
        'hard_filters': set<str>  # a non-empty subset of { ambiguous_sex,
            sex_aneuploidy,  low_coverage, bad_biallelic_metrics, contamination,
            chimera, coverage, insert_size }
    """
    logger.info('Generating hard filters')
    ht = mt.cols()
    ht = ht.annotate(hard_filters=hl.empty_set(hl.tstr))

    # Helper function to add filters into the `hard_filters` set
    def add_filter(ht, expr, name):
        return ht.annotate(
            hard_filters=hl.if_else(
                expr & hl.is_defined(expr), ht.hard_filters.add(name), ht.hard_filters
            )
        )

    # Remove samples with ambiguous sex assignments
    ht = add_filter(ht, sex_ht[ht.key].sex_karyotype == 'ambiguous', 'ambiguous_sex')
    ht = add_filter(
        ht,
        ~hl.set({'ambiguous', 'XX', 'XY'}).contains(sex_ht[ht.key].sex_karyotype),
        'sex_aneuploidy',
    )

    # Remove low-coverage samples
    # chrom 20 coverage is computed to infer sex and used here
    ht = add_filter(
        ht, sex_ht[ht.key].chr20_mean_dp < cutoffs_d['min_coverage'], 'low_coverage'
    )

    # Remove extreme raw bi-allelic sample QC outliers
    ht = add_filter(
        ht,
        (
            (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.n_snp
                > cutoffs_d['max_n_snps']
            )
            | (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.n_snp
                < cutoffs_d['min_n_snps']
            )
            | (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.n_singleton
                > cutoffs_d['max_n_singletons']
            )
            | (
                hail_sample_qc_ht[ht.key].bi_allelic_sample_qc.r_het_hom_var
                > cutoffs_d['max_r_het_hom']
            )
        ),
        'bad_biallelic_metrics',
    )

    ht = add_filter(
        ht,
        hl.is_defined(picard_metrics_ht[ht.key].r_contamination)
        & (
            picard_metrics_ht[ht.key].r_contamination > cutoffs_d['max_r_contamination']
        ),
        'contamination',
    )
    ht = add_filter(
        ht,
        hl.is_defined(picard_metrics_ht[ht.key].r_chimera)
        & (picard_metrics_ht[ht.key].r_chimera > cutoffs_d['max_r_chimera']),
        'chimera',
    )
    ht = add_filter(
        ht,
        hl.is_defined(picard_metrics_ht[ht.key].r_duplication)
        & (picard_metrics_ht[ht.key].r_duplication > cutoffs_d['max_r_duplication']),
        'dup_rate',
    )
    ht = add_filter(
        ht,
        hl.is_defined(picard_metrics_ht[ht.key].insert_size)
        & (picard_metrics_ht[ht.key].insert_size < cutoffs_d['min_insert_size']),
        'insert_size',
    )
    ht = ht.annotate_globals(hard_filter_cutoffs=hl.struct(**cutoffs_d))
    ht = ht.filter(hl.len(ht.hard_filters) > 0)
    return ht.checkpoint(out_ht_path, overwrite=True)


if __name__ == '__main__':
    main()  # pylint: disable=E1120
